import os
from logging import getLogger

from PyQt5.QtCore import Qt, QSettings, QTimer, QSize, pyqtSignal, pyqtSlot
from PyQt5.QtGui import QCloseEvent, QCursor
from PyQt5.QtWidgets import (
    QMainWindow,
    QApplication,
    QStatusBar,
    QScrollArea,
    QScroller,
    QComboBox,
    QMessageBox,
)

from rare.components.tabs import TabWidget
from rare.components.tray_icon import TrayIcon
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton, ArgumentsSingleton
from rare.utils.paths import lock_file

logger = getLogger("MainWindow")


class MainWindow(QMainWindow):
    # int: exit code
    exit_app: pyqtSignal = pyqtSignal(int)

    def __init__(self, parent=None):
        self._exit_code = 0
        self._accept_close = False
        self._window_launched = False
        super(MainWindow, self).__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.args = ArgumentsSingleton()

        self.settings = QSettings()

        self.setWindowTitle("Rare - GUI for legendary")
        self.tab_widget = TabWidget(self)
        self.tab_widget.exit_app.connect(self.on_exit_app)
        self.setCentralWidget(self.tab_widget)

        self.status_bar = QStatusBar()
        self.setStatusBar(self.status_bar)

        width, height = 1280, 720
        if self.settings.value("save_size", False, bool):
            width, height = self.settings.value("window_size", (width, height), tuple)

        self.resize(width, height)

        if not self.args.offline:
            try:
                from rare.utils.rpc import DiscordRPC

                self.rpc = DiscordRPC()
            except ModuleNotFoundError:
                logger.warning("Discord RPC module not found")

        self.timer = QTimer()
        self.timer.timeout.connect(self.timer_finished)
        self.timer.start(1000)

        self.tray_icon: TrayIcon = TrayIcon(self)
        self.tray_icon.exit_app.connect(self.on_exit_app)
        self.tray_icon.show_app.connect(self.show)
        self.tray_icon.activated.connect(lambda r: self.toggle() if r == self.tray_icon.DoubleClick else None)

        self.signals.send_notification.connect(
            lambda title: self.tray_icon.showMessage(
                self.tr("Download finished"),
                self.tr("Download finished. {} is playable now").format(title),
                self.tray_icon.Information,
                4000,
            )
            if self.settings.value("notification", True, bool)
            else None
        )

        # enable kinetic scrolling
        for scroll_area in self.findChildren(QScrollArea):
            if not scroll_area.property("no_kinetic_scroll"):
                QScroller.grabGesture(scroll_area.viewport(), QScroller.LeftMouseButtonGesture)

            # fix scrolling
            for combo_box in scroll_area.findChildren(QComboBox):
                combo_box.wheelEvent = lambda e: e.ignore()

    def center_window(self):
        # get the margins of the decorated window
        margins = self.windowHandle().frameMargins()
        # get the screen the cursor is on
        current_screen = QApplication.screenAt(QCursor.pos())
        if not current_screen:
            current_screen = QApplication.primaryScreen()
        # get the available screen geometry (excludes panels/docks)
        screen_rect = current_screen.availableGeometry()
        decor_width = margins.left() + margins.right()
        decor_height = margins.top() + margins.bottom()
        window_size = QSize(self.width(), self.height()).boundedTo(
            screen_rect.size() - QSize(decor_width, decor_height)
        )

        self.resize(window_size)
        self.move(screen_rect.center() - self.rect().adjusted(0, 0, decor_width, decor_height).center())

    @pyqtSlot()
    def show(self) -> None:
        super(MainWindow, self).show()
        if not self._window_launched:
            self.center_window()
        self._window_launched = True

    def hide(self) -> None:
        if self.settings.value("save_size", False, bool):
            size = self.size().width(), self.size().height()
            self.settings.setValue("window_size", size)
        super(MainWindow, self).hide()

    def toggle(self):
        if self.isHidden():
            self.show()
        else:
            self.hide()

    def timer_finished(self):
        file_path = lock_file()
        if os.path.exists(file_path):
            file = open(file_path, "r")
            action = file.read()
            file.close()
            if action.startswith("show"):
                self.show()
            os.remove(file_path)
        self.timer.start(1000)

    @pyqtSlot()
    @pyqtSlot(int)
    def on_exit_app(self, exit_code=0) -> None:
        self._exit_code = exit_code
        self.close()

    def close(self) -> bool:
        self._accept_close = True
        return super(MainWindow, self).close()

    def closeEvent(self, e: QCloseEvent) -> None:
        # lk: `accept_close` is set to `True` by the `close()` method, overrides exiting to tray in `closeEvent()`
        # lk: ensures exiting instead of hiding when `close()` is called programmatically
        if not self._accept_close:
            if self.settings.value("sys_tray", False, bool):
                self.hide()
                e.ignore()
                return
        # FIXME: Fix this with the download tab redesign
        if not self.args.offline and self.tab_widget.downloads_tab.is_download_active:
            reply = QMessageBox.question(
                self,
                self.tr("Quit {}?").format(QApplication.applicationName()),
                self.tr("There are active downloads. Are you sure you want to quit?"),
                buttons=(QMessageBox.Yes | QMessageBox.No),
                defaultButton=QMessageBox.No,
            )
            if reply == QMessageBox.Yes:
                # clear queue
                self.tab_widget.downloads_tab.queue_widget.update_queue([])
                self.tab_widget.downloads_tab.stop_download()
            else:
                e.ignore()
                return
        # FIXME: End of FIXME
        self.timer.stop()
        self.tray_icon.deleteLater()
        self.hide()
        self.exit_app.emit(self._exit_code)
        super(MainWindow, self).closeEvent(e)




import logging
import os
import shutil
import sys
import traceback
from argparse import Namespace
from datetime import datetime
from typing import Optional

import requests.exceptions
from PyQt5.QtCore import QThreadPool, QTimer, pyqtSlot, Qt
from PyQt5.QtWidgets import QApplication, QMessageBox
from requests import HTTPError

from rare.components.dialogs.launch_dialog import LaunchDialog
from rare.components.main_window import MainWindow
from rare.shared import (
    LegendaryCoreSingleton,
    GlobalSignalsSingleton,
    ArgumentsSingleton,
)
from rare.shared.rare_core import RareCore
from rare.utils import config_helper, paths
from rare.widgets.rare_app import RareApp

logger = logging.getLogger("Rare")


def excepthook(exc_type, exc_value, exc_tb):
    tb = "".join(traceback.format_exception(exc_type, exc_value, exc_tb))
    print("Error")
    if exc_tb == HTTPError:
        try:
            if LegendaryCoreSingleton().login():
                return
            else:
                raise ValueError
        except Exception as e:
            logger.fatal(str(e))
            QMessageBox.warning(None, "Error", QApplication.tr("Failed to login"))
            QApplication.exit(1)
            return
    logger.fatal(tb)
    QMessageBox.warning(None, "Error", tb)
    QApplication.exit(1)


class App(RareApp):
    def __init__(self, args: Namespace):
        log_file = "Rare_{0}.log"
        super(App, self).__init__(args, log_file)
        self.rare_core = RareCore(args=args)
        self.args = ArgumentsSingleton()
        self.signals = GlobalSignalsSingleton()
        self.core = LegendaryCoreSingleton()

        config_helper.init_config_handler(self.core)

        lang = self.settings.value("language", self.core.language_code, type=str)
        self.load_translator(lang)

        # set Application name for settings
        self.main_window: Optional[MainWindow] = None
        self.launch_dialog: Optional[LaunchDialog] = None
        self.timer: Optional[QTimer] = None

        # launch app
        self.launch_dialog = LaunchDialog(parent=None)
        self.launch_dialog.quit_app.connect(self.launch_dialog.close)
        self.launch_dialog.quit_app.connect(lambda x: sys.exit(x))
        self.launch_dialog.start_app.connect(self.start_app)
        self.launch_dialog.start_app.connect(self.launch_dialog.close)

        self.launch_dialog.login()

    def poke_timer(self):
        dt_exp = datetime.fromisoformat(self.core.lgd.userdata['expires_at'][:-1])
        dt_now = datetime.utcnow()
        td = abs(dt_exp - dt_now)
        self.timer.start(int(td.total_seconds() - 60) * 1000)
        logger.info(f"Renewed session expires at {self.core.lgd.userdata['expires_at']}")

    def re_login(self):
        logger.info("Session expires shortly. Renew session")
        try:
            self.core.login()
        except requests.exceptions.ConnectionError:
            self.timer.start(3000)  # try again if no connection
            return
        self.poke_timer()

    def start_app(self):
        self.timer = QTimer()
        self.timer.timeout.connect(self.re_login)
        self.poke_timer()

        self.main_window = MainWindow()
        self.main_window.exit_app.connect(self.on_exit_app)

        if not self.args.silent:
            self.main_window.show()

        if self.args.test_start:
            self.main_window.close()
            self.main_window = None
            self.on_exit_app(0)

    @pyqtSlot()
    @pyqtSlot(int)
    def on_exit_app(self, exit_code=0):
        threadpool = QThreadPool.globalInstance()
        threadpool.waitForDone()
        if self.timer is not None:
            self.timer.stop()
            self.timer.deleteLater()
            self.timer = None
        self.rare_core.deleteLater()
        del self.rare_core
        self.processEvents()
        shutil.rmtree(paths.tmp_dir())
        os.makedirs(paths.tmp_dir())

        self.exit(exit_code)


def start(args):
    # set excepthook to show dialog with exception
    sys.excepthook = excepthook

    while True:
        QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)
        QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)
        app = App(args)
        exit_code = app.exec_()
        # if not restart
        # restart app
        del app
        if exit_code != -133742:
            break





from logging import getLogger

from PyQt5.QtCore import Qt, QFileSystemWatcher
from PyQt5.QtWidgets import QGroupBox, QTableWidgetItem, QMessageBox, QPushButton

from rare.shared import LegendaryCoreSingleton
from rare.ui.components.tabs.settings.widgets.env_vars import Ui_EnvVars
from rare.utils import config_helper
from rare.utils.misc import icon

logger = getLogger("EnvVars")


class EnvVars(QGroupBox, Ui_EnvVars):
    def __init__(self, parent):
        super(EnvVars, self).__init__(parent=parent)
        self.setupUi(self)
        self.app_name = None
        self.core = LegendaryCoreSingleton()
        self.latest_item = None
        self.list_of_readonly = ["STEAM_COMPAT_DATA_PATH", "DXVK_HUD", "WINEPREFIX", "STEAM_COMPAT_CLIENT_INSTALL_PATH"]
        self.warn_msg = self.tr("Readonly, please edit this via the appropriate setting above.")
        self.setup_file_watcher()
        self.env_vars_table.cellChanged.connect(self.update_env_vars)
        self.env_vars_table.verticalHeader().sectionClicked.connect(self.remove_row)

    # We use this function to keep an eye on the config.
    # When the user uses for example the wineprefix settings, we need to update the table.
    # With this function, when the config file changes, we update the table.
    def setup_file_watcher(self):
        self.config_file_watcher = QFileSystemWatcher([str(self.core.lgd.config_path)], self)
        self.config_file_watcher.fileChanged.connect(self.import_env_vars)

    def append_row(self):
        # If the last row is not None, we insert a new one and set the correct icon.
        row_count = self.env_vars_table.rowCount()

        if row_count == 0:
            self.env_vars_table.insertRow(0)
            trash_icon = QTableWidgetItem()
            trash_icon.setIcon(icon("mdi.delete", "ei.minus"))
            self.env_vars_table.setVerticalHeaderItem(row_count, trash_icon)
            return

        last_item = self.env_vars_table.item(self.env_vars_table.rowCount() - 1, 0)

        if last_item is not None:
            self.env_vars_table.insertRow(row_count)
            trash_icon = QTableWidgetItem()
            trash_icon.setIcon(icon("mdi.delete", "ei.minus"))
            self.env_vars_table.setVerticalHeaderItem(row_count, trash_icon)

    def import_env_vars(self):
        self.env_vars_table.blockSignals(True)
        self.env_vars_table.clearContents()

        # If the config file doesnt have an env var section, we just set RowCount to 1 and return.
        if not self.core.lgd.config.has_section(f"{self.app_name}.env"):
            self.env_vars_table.setRowCount(1)
            trash_icon = QTableWidgetItem()
            trash_icon.setIcon(icon("mdi.delete", "ei.minus"))
            self.env_vars_table.setVerticalHeaderItem(0, trash_icon)
            self.env_vars_table.blockSignals(False)
            return

        # We count how many keys we have and insert new lines
        # (as many as we need).
        self.env_vars_table.setRowCount(len(self.core.lgd.config[f"{self.app_name}.env"]) + 1)

        # Each iteration we have to create a new QTableWidgetItem object,
        # else we segfault. (For using the same object in multiple references.)
        for i, (key, value) in enumerate(self.core.lgd.config[f"{self.app_name}.env"].items()):
            trash_icon = QTableWidgetItem()
            trash_icon.setIcon(icon("mdi.delete", "ei.minus"))
            self.env_vars_table.setVerticalHeaderItem(i, trash_icon)

            key_item = QTableWidgetItem()
            key_item.setText(key)
            self.env_vars_table.setItem(i, 0, key_item)

            value_item = QTableWidgetItem()
            value_item.setText(value)
            self.env_vars_table.setItem(i, 1, value_item)
            if key in self.list_of_readonly:
                key_item.setFlags(key_item.flags() ^ Qt.ItemIsEnabled)
                key_item.setToolTip(self.warn_msg)

                value_item.setFlags(value_item.flags() ^ Qt.ItemIsEnabled)
                value_item.setToolTip(self.warn_msg)

        trash_icon = QTableWidgetItem()
        trash_icon.setIcon(icon("mdi.delete", "ei.minus"))
        self.env_vars_table.setVerticalHeaderItem(self.env_vars_table.rowCount() - 1, trash_icon)

        self.env_vars_table.blockSignals(False)

    def update_env_vars(self, row, column):
        self.config_file_watcher.removePath(str(self.core.lgd.config_path))
        row_count = self.env_vars_table.rowCount()
        key_item = self.env_vars_table.item(row, 0)
        value_item = self.env_vars_table.item(row, 1)

        if key_item is not None and not key_item.text():
            try:
                list_of_config_keys = list(self.core.lgd.config[f"{self.app_name}.env"].keys())
            except KeyError:
                list_of_config_keys = []
            try:
                config_helper.remove_option(f"{self.app_name}.env", list_of_config_keys[row])
            except IndexError:
                # Item hasnt been saved to the config yet.
                pass
            self.env_vars_table.removeRow(key_item.row())
            self.append_row()
            return

        # get all config keys
        try:
            list_of_config_keys = list(self.core.lgd.config[f"{self.app_name}.env"].keys())
        except KeyError:
            list_of_config_keys = []

        # get all table keys
        list_of_keys_in_table = []
        for i in range(row_count):
            item = self.env_vars_table.item(i, 0)
            if item:
                list_of_keys_in_table.append(item.text())

        missing_item = list(set(list_of_config_keys) - set(list_of_keys_in_table))
        if len(missing_item) != 0:
            config_helper.remove_option(f"{self.app_name}.env", missing_item[0])

        # A env var always needs to have a key.
        # If it's none, we return.
        if key_item is None:
            return

        if key_item.text() in self.list_of_readonly:
            error_dialog = QMessageBox()
            error_dialog.setText(
                self.tr("Please don't manually add this environment variable. Use the appropriate game setting above."))
            error_dialog.exec()
            key_item.setText("")
            if value_item is not None:
                value_item.setText("")
            config_helper.remove_option(f"{self.app_name}.env", key_item.text())
            return

        if key_item.text():
            if "=" in key_item.text():
                error_dialog = QMessageBox()
                error_dialog.setText(
                    self.tr("Please don't use an equal sign in an env var."))
                error_dialog.exec()
                self.env_vars_table.removeRow(row)
                self.append_row()
                return

            if key_item.text() in list_of_config_keys and column == 0:
                ask_user = QMessageBox()
                ask_user.setText(
                    self.tr("The config already contains this environment variable."))
                ask_user.setInformativeText(
                    self.tr("Do you want to keep the newer one or the older one?"))

                ask_user.addButton(QPushButton("Keep the newer one"), QMessageBox.YesRole)
                ask_user.addButton(QPushButton("Keep the older one"), QMessageBox.NoRole)

                response = ask_user.exec()

                if response == 0:
                    if value_item is not None:
                        config_helper.add_option(f"{self.app_name}.env", key_item.text(), value_item.text())
                    else:
                        config_helper.add_option(f"{self.app_name}.env", key_item.text(), "")

                    item_to_safe = self.env_vars_table.findItems(key_item.text(), Qt.MatchExactly)

                    # aznd:
                    # This is to fix an issue where the user updates a env var, thats above the older one.
                    # so say if you have two env vars:
                    # something = newkey
                    # yes = oldkey
                    # if the user updates the something key to yes, it would delete the wrong row,
                    # since item_to_safe[0] is the first search result. but we dont want to delete that,
                    # we want to delete the second result.
                    # we use this simple if check to find out which case we have.
                    if key_item.row() < item_to_safe[0].row():
                        self.env_vars_table.removeRow(item_to_safe[0].row())
                    elif key_item.row() > item_to_safe[0].row():
                        self.env_vars_table.removeRow(item_to_safe[0].row())
                    else:
                        self.env_vars_table.removeRow(item_to_safe[1].row())

                    self.append_row()
                    return

                elif response == 1:
                    self.env_vars_table.removeRow(row)
                    self.append_row()
                    return

            # When the value_item is None, we just use an empty string for the value.
            if value_item is None:
                if self.latest_item in list_of_config_keys:
                    config_helper.remove_option(f"{self.app_name}.env", self.latest_item)
                    config_helper.save_config()

                config_helper.add_option(f"{self.app_name}.env", key_item.text(), "")
                config_helper.save_config()
            else:
                if self.latest_item in list_of_config_keys:
                    config_helper.remove_option(f"{self.app_name}.env", self.latest_item)
                    config_helper.save_config()

                config_helper.add_option(
                    f"{self.app_name}.env",
                    key_item.text(),
                    value_item.text()
                )
                config_helper.save_config()

        self.append_row()
        self.config_file_watcher.addPath(str(self.core.lgd.config_path))

    def remove_row(self, index):
        self.config_file_watcher.removePath(str(self.core.lgd.config_path))
        key_item = self.env_vars_table.item(index, 0)
        value_item = self.env_vars_table.item(index, 1)

        if key_item is None:
            if value_item is not None:
                value_item.setText("")
            return

        # If the user tries to delete one of the readonly vars, we return immediately.
        if key_item.text() in self.list_of_readonly:
            return

        if key_item is not None:
            self.env_vars_table.removeRow(index)
        try:
            list_of_keys = []
            for key in self.core.lgd.config[f"{self.app_name}.env"]:
                list_of_keys.append(key)
            config_helper.remove_option(f"{self.app_name}.env", list_of_keys[index])
        except (KeyError, IndexError):
            pass
        self.config_file_watcher.addPath(str(self.core.lgd.config_path))

    def check_if_item(self, item: QTableWidgetItem) -> bool:
        item_to_check = self.env_vars_table.findItems(item.text(), Qt.MatchExactly)
        if item_to_check[0].column() == 0:
            return False
        return True

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_Delete or e.key() == Qt.Key_Backspace:
            selected_items = self.env_vars_table.selectedItems()

            if len(selected_items) == 0:
                return

            item_in_table = self.env_vars_table.findItems(selected_items[0].text(), Qt.MatchExactly)

            # Our first selection is in column 0.  So, we have to find out if the user 
            # only selected keys, or keys and values. we use the check_if_item func
            if item_in_table[0].column() == 0:
                which_index_to_use = 1
                if len(selected_items) == 1:
                    which_index_to_use = 0
                if self.check_if_item(selected_items[which_index_to_use]):
                    # User selected keys and values, so we skip the values
                    for i in selected_items[::2]:
                        if i:
                            config_helper.remove_option(f"{self.app_name}.env", i.text())
                            self.env_vars_table.removeRow(i.row())
                    self.append_row()
                else:
                    # user only selected keys
                    for i in selected_items:
                        if i:
                            config_helper.remove_option(f"{self.app_name}.env", i.text())
                            self.env_vars_table.removeRow(i.row())
                    self.append_row()

            # User only selected values, so we just set the text to ""
            elif item_in_table[0].column() == 1:
                [i.setText("") for i in selected_items]

        elif e.key() == Qt.Key_Escape:
            e.ignore()

    def update_game(self, app_name):
        self.app_name = app_name
        self.import_env_vars()



import os
import platform
from logging import getLogger

from PyQt5.QtCore import Qt, QThreadPool
from PyQt5.QtWidgets import QSizePolicy, QPushButton, QLabel, QFileDialog, QMessageBox
from legendary.models.game import Game, InstalledGame

from rare.components.tabs.settings import DefaultGameSettings
from rare.components.tabs.settings.widgets.pre_launch import PreLaunchSettings
from rare.utils import config_helper
from rare.utils.extra_widgets import PathEdit
from rare.utils.misc import icon, WineResolver, get_raw_save_path

logger = getLogger("GameSettings")


class GameSettings(DefaultGameSettings):
    game: Game
    igame: InstalledGame

    def __init__(self, parent=None):
        super(GameSettings, self).__init__(False, parent)
        self.pre_launch_settings = PreLaunchSettings()
        self.launch_settings_group.layout().addRow(
            QLabel(self.tr("Pre launch command")), self.pre_launch_settings
        )

        self.cloud_save_path_edit = PathEdit(
            "",
            file_type=QFileDialog.DirectoryOnly,
            placeholder=self.tr("Cloud save path"),
            edit_func=lambda text: (True, text, None)
            if os.path.exists(text)
            else (False, text, PathEdit.reasons.dir_not_exist),
            save_func=self.save_save_path,
        )
        self.cloud_layout.addRow(QLabel(self.tr("Save path")), self.cloud_save_path_edit)

        self.compute_save_path_button = QPushButton(icon("fa.magic"), self.tr("Auto compute save path"))
        self.compute_save_path_button.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Fixed)
        self.compute_save_path_button.clicked.connect(self.compute_save_path)
        self.cloud_layout.addRow(None, self.compute_save_path_button)

        self.offline.currentIndexChanged.connect(lambda x: self.update_combobox(x, "offline"))
        self.skip_update.currentIndexChanged.connect(lambda x: self.update_combobox(x, "skip_update_check"))
        self.cloud_sync.stateChanged.connect(
            lambda: self.settings.setValue(
                f"{self.game.app_name}/auto_sync_cloud", self.cloud_sync.isChecked()
            )
        )
        self.override_exe_edit.textChanged.connect(lambda text: self.save_line_edit("override_exe", text))
        self.launch_params.textChanged.connect(lambda x: self.save_line_edit("start_params", x))

        self.game_settings_layout.setAlignment(Qt.AlignTop)

    def compute_save_path(self):
        if self.core.is_installed(self.game.app_name) and self.game.supports_cloud_saves:
            try:
                new_path = self.core.get_save_path(self.game.app_name)
            except Exception as e:
                logger.warning(str(e))
                resolver = WineResolver(get_raw_save_path(self.game), self.game.app_name)
                if not resolver.wine_env.get("WINEPREFIX"):
                    self.cloud_save_path_edit.setText("")
                    QMessageBox.warning(self, "Warning", "No wine prefix selected. Please set it in settings")
                    return
                self.cloud_save_path_edit.setText(self.tr("Loading"))
                self.cloud_save_path_edit.setDisabled(True)
                self.compute_save_path_button.setDisabled(True)

                app_name = self.game.app_name[:]
                resolver.signals.result_ready.connect(lambda x: self.wine_resolver_finished(x, app_name))
                QThreadPool.globalInstance().start(resolver)
                return
            else:
                self.cloud_save_path_edit.setText(new_path)

    def wine_resolver_finished(self, path, app_name):
        logger.info(f"Wine resolver finished for {app_name}. Computed save path: {path}")
        if app_name == self.game.app_name:
            self.cloud_save_path_edit.setDisabled(False)
            self.compute_save_path_button.setDisabled(False)
            if path and not os.path.exists(path):
                try:
                    os.makedirs(path)
                except PermissionError:
                    self.cloud_save_path_edit.setText("")
                    QMessageBox.warning(
                        None,
                        "Error",
                        self.tr("Error while launching {}. No permission to create {}").format(
                            self.game.app_title, path
                        ),
                    )
                    return
            if not path:
                self.cloud_save_path_edit.setText("")
                return
            self.cloud_save_path_edit.setText(path)
        elif path:
            igame = self.core.get_installed_game(app_name)
            igame.save_path = path
            self.core.lgd.set_installed_game(app_name, igame)

    def save_save_path(self, text):
        if self.game.supports_cloud_saves and self.change:
            self.igame.save_path = text
            self.core.lgd.set_installed_game(self.igame.app_name, self.igame)

    def save_line_edit(self, option, value):
        if value:

            config_helper.add_option(self.game.app_name, option, value)
        else:
            config_helper.remove_option(self.game.app_name, option)
        config_helper.save_config()

        if option == "wine_prefix":
            if self.game.supports_cloud_saves:
                self.compute_save_path()

    def update_combobox(self, i, option):
        if self.change:
            # remove section
            if i:
                if i == 1:
                    config_helper.add_option(self.game.app_name, option, "true")
                if i == 2:
                    config_helper.add_option(self.game.app_name, option, "false")
            else:
                config_helper.remove_option(self.game.app_name, option)
            config_helper.save_config()

    def load_settings(self, app_name):
        self.change = False
        super(GameSettings, self).load_settings(app_name)
        self.game = self.core.get_game(app_name)
        self.igame = self.core.get_installed_game(self.game.app_name)
        if self.igame:
            if self.igame.can_run_offline:
                offline = self.core.lgd.config.get(self.game.app_name, "offline", fallback="unset")
                if offline == "true":
                    self.offline.setCurrentIndex(1)
                elif offline == "false":
                    self.offline.setCurrentIndex(2)
                else:
                    self.offline.setCurrentIndex(0)

                self.offline.setEnabled(True)
            else:
                self.offline.setEnabled(False)
        else:
            self.offline.setEnabled(False)

        skip_update = self.core.lgd.config.get(self.game.app_name, "skip_update_check", fallback="unset")
        if skip_update == "true":
            self.skip_update.setCurrentIndex(1)
        elif skip_update == "false":
            self.skip_update.setCurrentIndex(2)
        else:
            self.skip_update.setCurrentIndex(0)

        self.title.setTitle(self.game.app_title)
        if platform.system() != "Windows":
            if self.igame and self.igame.platform == "Mac":
                self.linux_settings_widget.setVisible(False)
            else:
                self.linux_settings_widget.setVisible(True)

        if not self.game.supports_cloud_saves:
            self.cloud_group.setEnabled(False)
            self.cloud_save_path_edit.setText("")
        else:
            self.cloud_group.setEnabled(True)
            sync_cloud = self.settings.value(f"{self.game.app_name}/auto_sync_cloud", True, bool)
            self.cloud_sync.setChecked(sync_cloud)
            if self.igame.save_path:
                self.cloud_save_path_edit.setText(self.igame.save_path)
            else:
                self.cloud_save_path_edit.setText("")

        self.launch_params.setText(self.core.lgd.config.get(self.game.app_name, "start_params", fallback=""))
        self.override_exe_edit.setText(
            self.core.lgd.config.get(self.game.app_name, "override_exe", fallback="")
        )

        self.pre_launch_settings.load_settings(app_name)

        self.change = True



import os
import platform
from logging import getLogger

from PyQt5.QtCore import pyqtSignal, QObject, QRunnable, QStandardPaths
from legendary.core import LegendaryCore

from rare.lgndr.cli import LegendaryCLI
from rare.lgndr.glue.arguments import LgndrVerifyGameArgs, LgndrUninstallGameArgs
from rare.lgndr.glue.monkeys import LgndrIndirectStatus
from rare.shared import LegendaryCoreSingleton, ArgumentsSingleton
from rare.utils import config_helper

logger = getLogger("Legendary Utils")


def uninstall_game(core: LegendaryCore, app_name: str, keep_files=False, keep_config=False):
    igame = core.get_installed_game(app_name)

    # remove shortcuts link
    desktop = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
    applications = QStandardPaths.writableLocation(QStandardPaths.ApplicationsLocation)
    if platform.system() == "Linux":
        desktop_shortcut = os.path.join(desktop, f"{igame.title}.desktop")
        if os.path.exists(desktop_shortcut):
            os.remove(desktop_shortcut)

        applications_shortcut = os.path.join(applications, f"{igame.title}.desktop")
        if os.path.exists(applications_shortcut):
            os.remove(applications_shortcut)

    elif platform.system() == "Windows":
        game_title = igame.title.split(":")[0]
        desktop_shortcut = os.path.join(desktop, f"{game_title}.lnk")
        if os.path.exists(desktop_shortcut):
            os.remove(desktop_shortcut)

        start_menu_shortcut = os.path.join(applications, "..", f"{game_title}.lnk")
        if os.path.exists(start_menu_shortcut):
            os.remove(start_menu_shortcut)

    status = LgndrIndirectStatus()
    LegendaryCLI(core).uninstall_game(
        LgndrUninstallGameArgs(
            app_name=app_name,
            keep_files=keep_files,
            indirect_status=status,
            yes=True,
        )
    )
    if not keep_config:
        logger.info("Removing sections in config file")
        config_helper.remove_section(app_name)
        config_helper.remove_section(f"{app_name}.env")

        config_helper.save_config()

    return status.success, status.message


def update_manifest(app_name: str, core: LegendaryCore):
    game = core.get_game(app_name)
    logger.info(f"Reloading game manifest of {game.app_title}")
    new_manifest_data, base_urls = core.get_cdn_manifest(game)
    # overwrite base urls in metadata with current ones to avoid using old/dead CDNs
    game.base_urls = base_urls
    # save base urls to game metadata
    core.lgd.set_game_meta(game.app_name, game)

    new_manifest = core.load_manifest(new_manifest_data)
    logger.debug(f"Base urls: {base_urls}")
    # save manifest with version name as well for testing/downgrading/etc.
    core.lgd.save_manifest(game.app_name, new_manifest_data, version=new_manifest.meta.build_version)


class VerifyWorker(QRunnable):
    class Signals(QObject):
        status = pyqtSignal(str, int, int, float, float)
        result = pyqtSignal(str, bool, int, int)
        error = pyqtSignal(str, str)

    num: int = 0
    total: int = 1  # set default to 1 to avoid DivisionByZero before it is initialized

    def __init__(self, app_name):
        super(VerifyWorker, self).__init__()
        self.signals = VerifyWorker.Signals()
        self.setAutoDelete(True)
        self.core = LegendaryCoreSingleton()
        self.args = ArgumentsSingleton()
        self.app_name = app_name

    def status_callback(self, num: int, total: int, percentage: float, speed: float):
        self.signals.status.emit(self.app_name, num, total, percentage, speed)

    def run(self):
        cli = LegendaryCLI(self.core)
        status = LgndrIndirectStatus()
        args = LgndrVerifyGameArgs(
            app_name=self.app_name, indirect_status=status, verify_stdout=self.status_callback
        )

        # lk: first pass, verify with the current manifest
        repair_mode = False
        result = cli.verify_game(
            args, print_command=False, repair_mode=repair_mode, repair_online=not self.args.offline
        )
        if result is None:
            # lk: second pass with downloading the latest manifest
            # lk: this happens if the manifest was not found and repair_mode was not requested
            # lk: we already have checked if the directory exists before starting the worker
            try:
                # lk: this try-except block handles the exception caused by a missing manifest
                # lk: and is raised only in the case we are offline
                repair_mode = True
                result = cli.verify_game(
                    args, print_command=False, repair_mode=repair_mode, repair_online=not self.args.offline
                )
                if result is None:
                    raise ValueError
            except ValueError:
                self.signals.error.emit(self.app_name, status.message)
                return

        success = result is not None and not any(result)
        if success:
            # lk: if verification was successful we delete the repair file and run the clean procedure
            # lk: this could probably be cut down to what is relevant for this use-case and skip the `cli` call
            igame = self.core.get_installed_game(self.app_name)
            game = self.core.get_game(self.app_name, platform=igame.platform)
            repair_file = os.path.join(self.core.lgd.get_tmp_path(), f"{self.app_name}.repair")
            cli.install_game_cleanup(game=game, igame=igame, repair_mode=True, repair_file=repair_file)

        self.signals.result.emit(self.app_name, success, *result)



from PyQt5.QtCore import QObject, pyqtSignal

from .install import InstallOptionsModel


class GlobalSignals(QObject):
    exit_app = pyqtSignal(int)  # exit code
    send_notification = pyqtSignal(str)  # app_title

    set_main_tab_index = pyqtSignal(int)  # tab index
    update_download_tab_text = pyqtSignal()

    dl_progress = pyqtSignal(int)  # 0-100
    # set visibility of installing widget in games tab
    installation_started = pyqtSignal(str)  # app_name
    add_download = pyqtSignal(str)

    install_game = pyqtSignal(InstallOptionsModel)
    installation_finished = pyqtSignal(bool, str)

    overlay_installation_finished = pyqtSignal()

    update_gamelist = pyqtSignal(list)
    game_uninstalled = pyqtSignal(str)  # appname

    set_discord_rpc = pyqtSignal(str)  # app_name of running game
    rpc_settings_updated = pyqtSignal()

    wine_prefix_updated = pyqtSignal()



import logging
from dataclasses import dataclass

from typing_extensions import Protocol


class GetBooleanChoiceProtocol(Protocol):
    def __call__(self, prompt: str, default: bool = ...) -> bool:
        ...


def get_boolean_choice(prompt: str, default: bool = True) -> bool:
    return default


def verify_stdout(a0: int, a1: int, a2: float, a3: float) -> None:
    print(f"Verification progress: {a0}/{a1} ({a2:.01f}%) [{a3:.1f} MiB/s]\t\r")


class DLManagerSignals:
    _kill = False
    _update = False

    @property
    def kill(self) -> bool:
        self._update = False
        return self._kill

    @kill.setter
    def kill(self, value: bool) -> None:
        self._update = True
        self._kill = value

    @property
    def update(self) -> bool:
        _update = self._update
        self._update = False
        return _update


@dataclass
class LgndrIndirectStatus:
    success: bool = False
    message: str = ""

    def __len__(self):
        if self.message:
            return 2
        else:
            return 0

    def __bool__(self):
        return self.success

    def __getitem__(self, item):
        if item == 0:
            return self.success
        elif item == 1:
            return self.message
        else:
            raise IndexError

    def __iter__(self):
        return iter((self.success, self.message))

    def __str__(self):
        return self.message


class LgndrIndirectLogger:
    def __init__(
        self, status: LgndrIndirectStatus, logger: logging.Logger = None, level: int = logging.ERROR
    ):
        self.logger = logger
        self.level = level
        self.status = status

    def set_logger(self, logger: logging.Logger):
        self.logger = logger

    def set_level(self, level: int):
        self.level = level

    def _log(self, level: int, msg: str):
        self.status.success = True if level < self.level else False
        self.status.message = msg
        if self.logger:
            self.logger.log(level, msg)

    def debug(self, msg: str):
        self._log(logging.DEBUG, msg)

    def info(self, msg: str):
        self._log(logging.INFO, msg)

    def warning(self, msg: str):
        self._log(logging.WARNING, msg)

    def error(self, msg: str):
        self._log(logging.ERROR, msg)

    def critical(self, msg: str):
        self._log(logging.CRITICAL, msg)

    def fatal(self, msg: str):
        self.critical(msg)



import hashlib
import json
import pickle
import zlib
# from concurrent import futures
from logging import getLogger
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Tuple, Dict, Union, Type, List, Callable

import requests
from PyQt5.QtCore import (
    Qt,
    pyqtSignal,
    QObject,
    QSize,
    QThreadPool,
    QRunnable,
)
from PyQt5.QtGui import (
    QPixmap,
    QImage,
    QPainter,
)
from PyQt5.QtWidgets import QApplication
from legendary.models.game import Game

from rare.lgndr.core import LegendaryCore
from rare.models.signals import GlobalSignals
from rare.utils.paths import image_dir, resources_path

# from requests_futures.sessions import FuturesSession

if TYPE_CHECKING:
    pass

logger = getLogger("ImageManager")


class ImageSize:
    class Preset:
        def __init__(self, divisor: float, pixel_ratio: float):
            self.__img_factor = 67
            self.__divisor = divisor
            self.__pixel_ratio = pixel_ratio
            self.__size = QSize(self.__img_factor * 3, self.__img_factor * 4) * pixel_ratio / divisor
            # lk: for prettier images set this to true
            self.__smooth_transform: bool = False
            if divisor > 2:
                self.__smooth_transform = False

        @property
        def size(self) -> QSize:
            return self.__size

        @property
        def divisor(self) -> float:
            return self.__divisor

        @property
        def smooth(self) -> bool:
            return self.__smooth_transform

        @property
        def pixel_ratio(self) -> float:
            return self.__pixel_ratio

    Image = Preset(1, 2)
    """! @brief Size and pixel ratio of the image on disk"""

    Display = Preset(1, 1)
    """! @brief Size and pixel ratio for displaying"""

    Normal = Display
    """! @brief Same as Display"""

    Small = Preset(3, 1)
    """! @brief Small image size for displaying"""

    Smaller = Preset(4, 1)
    """! @brief Smaller image size for displaying"""

    Icon = Preset(5, 1)
    """! @brief Smaller image size for UI icons"""


class ImageManager(QObject):
    class Worker(QRunnable):
        class Signals(QObject):
            # object: Game
            completed = pyqtSignal(object)

        def __init__(self, func: Callable, updates: List, json_data: Dict, game: Game):
            super(ImageManager.Worker, self).__init__()
            self.signals = ImageManager.Worker.Signals()
            self.setAutoDelete(True)
            self.func = func
            self.updates = updates
            self.json_data = json_data
            self.game = game

        def run(self):
            self.func(self.updates, self.json_data, self.game)
            logger.debug(f" Emitting singal for game {self.game.app_name} - {self.game.app_title}")
            self.signals.completed.emit(self.game)

    def __init__(self, signals: GlobalSignals, core: LegendaryCore):
        # lk: the ordering in __img_types matters for the order of fallbacks
        self.__img_types: Tuple = ("DieselGameBoxTall", "Thumbnail", "DieselGameBoxLogo")
        self.__dl_retries = 1
        self.__worker_app_names: List[str] = []
        super(QObject, self).__init__()
        self.signals = signals
        self.core = core

        self.image_dir: Path = image_dir()
        if not self.image_dir.is_dir():
            self.image_dir.mkdir()
            logger.info(f"Created image directory at {self.image_dir}")

        self.device = ImageSize.Preset(1, QApplication.instance().devicePixelRatio())

        self.threadpool = QThreadPool()
        self.threadpool.setMaxThreadCount(8)

    def __img_dir(self, app_name: str) -> Path:
        return self.image_dir.joinpath(app_name)

    def __img_json(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("image.json")

    def __img_cache(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("image.cache")

    def __img_color(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("installed.png")

    def __img_gray(self, app_name: str) -> Path:
        return self.__img_dir(app_name).joinpath("uninstalled.png")

    def __prepare_download(self, game: Game, force: bool = False) -> Tuple[List, Dict]:
        if force and self.__img_dir(game.app_name).exists():
            self.__img_color(game.app_name).unlink(missing_ok=True)
            self.__img_color(game.app_name).unlink(missing_ok=True)
        if not self.__img_dir(game.app_name).is_dir():
            self.__img_dir(game.app_name).mkdir()

        # Load image checksums
        if not self.__img_json(game.app_name).is_file():
            json_data: Dict = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        else:
            json_data = json.load(open(self.__img_json(game.app_name), "r"))

        # lk: fast path for games without images, convert Rare's logo
        if not game.metadata.get("keyImages", False):
            if not self.__img_color(game.app_name).is_file() or not self.__img_gray(game.app_name).is_file():
                cache_data: Dict = dict(zip(self.__img_types, [None] * len(self.__img_types)))
                cache_data["DieselGameBoxTall"] = open(
                    resources_path.joinpath("images", "cover.png"), "rb"
                ).read()
                # cache_data["DieselGameBoxLogo"] = open(
                #         resources_path.joinpath("images", "Rare_nonsquared.png"), "rb").read()
                self.__convert(game, cache_data)
                json_data["cache"] = None
                json_data["scale"] = ImageSize.Image.pixel_ratio
                json_data["size"] = ImageSize.Image.size.__str__()
                json.dump(json_data, open(self.__img_json(game.app_name), "w"))

        # lk: Find updates or initialize if images are missing.
        # lk: `updates` will be empty for games without images
        # lk: so everything below it is skipped
        if not self.__img_color(game.app_name).is_file() or not self.__img_gray(game.app_name).is_file():
            updates = [image for image in game.metadata["keyImages"] if image["type"] in self.__img_types]
        else:
            updates = list()
            for image in game.metadata["keyImages"]:
                if image["type"] in self.__img_types:
                    if json_data[image["type"]] != image["md5"]:
                        updates.append(image)

        return updates, json_data

    def __download(self, updates, json_data, game, use_async: bool = False) -> bool:
        # Decompress existing image.cache
        if not self.__img_cache(game.app_name).is_file():
            cache_data = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        else:
            cache_data = self.__decompress(game)

        # lk: filter updates again against the cache now that it is available
        updates = [
            image
            for image in updates
            if cache_data[image["type"]] is None or json_data[image["type"]] != image["md5"]
        ]

        # Download
        # # lk: Keep this so I don't have to go looking for it again,
        # # lk: it might be useful in the future.
        # if use_async and len(updates) > 1:
        #     session = FuturesSession(max_workers=len(self.__img_types))
        #     image_requests = []
        #     for image in updates:
        #         logger.info(f"Downloading {image['type']} for {game.app_title}")
        #         json_data[image["type"]] = image["md5"]
        #         payload = {"resize": 1, "w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}
        #         req = session.get(image["url"], params=payload)
        #         req.image_type = image["type"]
        #         image_requests.append(req)
        #     for req in futures.as_completed(image_requests):
        #         cache_data[req.image_type] = req.result().content
        # else:
        for image in updates:
            logger.info(f"Downloading {image['type']} for {game.app_title}")
            json_data[image["type"]] = image["md5"]
            payload = {"resize": 1, "w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}
            cache_data[image["type"]] = requests.get(image["url"], params=payload).content

        self.__convert(game, cache_data)
        # lk: don't keep the cache if there is no logo (kept for me)
        # if cache_data["DieselGameBoxLogo"] is not None:
        #     self.__compress(game, cache_data)
        self.__compress(game, cache_data)

        # hash image cache
        try:
            with open(self.__img_cache(game.app_name), "rb") as archive:
                archive_hash = hashlib.md5(archive.read()).hexdigest()
        except FileNotFoundError:
            archive_hash = None

        json_data["cache"] = archive_hash
        json_data["scale"] = ImageSize.Image.pixel_ratio
        json_data["size"] = {"w": ImageSize.Image.size.width(), "h": ImageSize.Image.size.height()}

        # write image.json
        with open(self.__img_json(game.app_name), "w") as file:
            json.dump(json_data, file)

        return bool(updates)

    def __convert(self, game, images, force=False) -> None:
        for image in [self.__img_color(game.app_name), self.__img_gray(game.app_name)]:
            if force and image.exists():
                image.unlink(missing_ok=True)

        cover_data = None
        for image_type in self.__img_types:
            if images[image_type] is not None:
                cover_data = images[image_type]
                break

        cover = QImage()
        cover.loadFromData(cover_data)
        cover.convertToFormat(QImage.Format_ARGB32_Premultiplied)
        # lk: Images are not always 4/3, crop them to size
        factor = min(cover.width() // 3, cover.height() // 4)
        rem_w = (cover.width() - factor * 3) // 2
        rem_h = (cover.height() - factor * 4) // 2
        cover = cover.copy(rem_w, rem_h, factor * 3, factor * 4)

        if images["DieselGameBoxLogo"] is not None:
            logo = QImage()
            logo.loadFromData(images["DieselGameBoxLogo"])
            logo.convertToFormat(QImage.Format_ARGB32_Premultiplied)
            if logo.width() > cover.width():
                logo = logo.scaled(cover.width(), cover.height(), Qt.KeepAspectRatio, Qt.SmoothTransformation)
            painter = QPainter(cover)
            painter.drawImage((cover.width() - logo.width()) // 2, cover.height() - logo.height(), logo)
            painter.end()

        cover = cover.scaled(ImageSize.Image.size, Qt.KeepAspectRatio, Qt.SmoothTransformation)

        # this is not required if we ever want to re-apply the alpha channel
        # cover = cover.convertToFormat(QImage.Format_Indexed8)

        # add the alpha channel back to the cover
        cover = cover.convertToFormat(QImage.Format_ARGB32_Premultiplied)

        cover.save(
            str(self.__img_color(game.app_name)),
            format="PNG",
        )
        # quick way to convert to grayscale
        cover = cover.convertToFormat(QImage.Format_Grayscale8)
        # add the alpha channel back to the grayscale cover
        cover = cover.convertToFormat(QImage.Format_ARGB32_Premultiplied)
        cover.save(
            str(self.__img_gray(game.app_name)),
            format="PNG",
        )

    def __compress(self, game: Game, data: Dict) -> None:
        archive = open(self.__img_cache(game.app_name), "wb")
        cdata = zlib.compress(pickle.dumps(data), level=-1)
        archive.write(cdata)
        archive.close()

    def __decompress(self, game: Game) -> Dict:
        archive = open(self.__img_cache(game.app_name), "rb")
        try:
            data = zlib.decompress(archive.read())
            data = pickle.loads(data)
        except zlib.error:
            data = dict(zip(self.__img_types, [None] * len(self.__img_types)))
        finally:
            archive.close()
        return data

    def download_image(
        self, game: Game, load_callback: Callable[[Game], None], priority: int, force: bool = False
    ) -> None:
        updates, json_data = self.__prepare_download(game, force)
        if not updates:
            load_callback(game)
            return
        if updates and game.app_name not in self.__worker_app_names:
            image_worker = ImageManager.Worker(self.__download, updates, json_data, game)
            self.__worker_app_names.append(game.app_name)

            image_worker.signals.completed.connect(load_callback)
            image_worker.signals.completed.connect(lambda g: self.__worker_app_names.remove(g.app_name))
            self.threadpool.start(image_worker, priority)

    def download_image_blocking(self, game: Game, force: bool = False) -> None:
        updates, json_data = self.__prepare_download(game, force)
        if not updates:
            return
        if updates:
            self.__download(updates, json_data, game, use_async=True)

    def __get_cover(
        self, container: Union[Type[QPixmap], Type[QImage]], app_name: str, color: bool = True
    ) -> Union[QPixmap, QImage]:
        ret = container()
        if not app_name:
            raise RuntimeError("app_name is an empty string")
        if color:
            if self.__img_color(app_name).is_file():
                ret.load(str(self.__img_color(app_name)))
        else:
            if self.__img_gray(app_name).is_file():
                ret.load(str(self.__img_gray(app_name)))
        if not ret.isNull():
            ret.setDevicePixelRatio(ImageSize.Image.pixel_ratio)
            # lk: Scaling happens at painting. It might be inefficient so leave this here as an alternative
            # lk: If this is uncommented, the transformation in ImageWidget should be adjusted also
            ret = ret.scaled(self.device.size, Qt.KeepAspectRatio, Qt.SmoothTransformation)
            ret.setDevicePixelRatio(self.device.pixel_ratio)
        return ret

    def get_pixmap(self, app_name: str, color: bool = True) -> QPixmap:
        """
        Use when the image is to be presented directly on the screen.

        @param app_name: The RareGame object for this game
        @param color: True to load the colored pixmap, False to load the grayscale
        @return: QPixmap
        """
        pixmap: QPixmap = self.__get_cover(QPixmap, app_name, color)
        return pixmap

    def get_image(self, app_name: str, color: bool = True) -> QImage:
        """
        Use when the image has to be manipulated before being rendered.

        @param app_name: The RareGame object for this game
        @param color: True to load the colored image, False to load the grayscale
        @return: QImage
        """
        image: QImage = self.__get_cover(QImage, app_name, color)
        return image



import datetime
import sys
from dataclasses import dataclass
from logging import getLogger
from typing import Union, List, Dict

from PyQt5.QtCore import QObject, pyqtSignal, QRunnable, QThreadPool, Qt, QSettings
from PyQt5.QtWidgets import QDialog, QMessageBox, QSizePolicy, QLayout, QApplication

from legendary.core import LegendaryCore
from legendary.models.game import SaveGameStatus, InstalledGame, SaveGameFile
from rare.shared import LegendaryCoreSingleton, ArgumentsSingleton, ApiResultsSingleton
from rare.ui.components.dialogs.sync_save_dialog import Ui_SyncSaveDialog
from rare.utils.misc import icon

logger = getLogger("Cloud Saves")


@dataclass
class UploadModel:
    app_name: str
    date_time: datetime.datetime
    path: str


@dataclass
class DownloadModel:
    app_name: str
    latest_save: SaveGameFile
    path: str


class SaveSignals(QObject):
    finished = pyqtSignal(str, str)


class SaveWorker(QRunnable):
    def __init__(self, model: Union[UploadModel, DownloadModel]):
        super(SaveWorker, self).__init__()
        self.signals = SaveSignals()
        self.setAutoDelete(True)
        self.core = LegendaryCoreSingleton()
        self.api_results = ApiResultsSingleton()
        self.model = model

    def run(self) -> None:
        try:
            if isinstance(self.model, DownloadModel):
                self.core.download_saves(
                    self.model.app_name,
                    self.model.latest_save.manifest_name,
                    self.model.path,
                )
            else:
                self.core.upload_save(
                    self.model.app_name, self.model.path, self.model.date_time
                )
        except Exception as e:
            self.signals.finished.emit(str(e), self.model.app_name)
            logger.error(str(e))
            return
        try:
            if isinstance(self.model, UploadModel):
                logger.info("Updating cloud saves...")
                result = self.core.get_save_games(self.model.app_name)
                self.api_results.saves = result
        except Exception as e:
            self.signals.finished.emit(str(e), self.model.app_name)
            logger.error(str(e))
            return

        self.signals.finished.emit("", self.model.app_name)


class CloudSaveDialog(QDialog, Ui_SyncSaveDialog):
    DOWNLOAD = 2
    UPLOAD = 1
    CANCEL = 0

    def __init__(
            self,
            igame: InstalledGame,
            dt_local: datetime.datetime,
            dt_remote: datetime.datetime,
            newer: str,
    ):
        super(CloudSaveDialog, self).__init__()
        self.setupUi(self)

        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setWindowFlags(Qt.Dialog | Qt.CustomizeWindowHint | Qt.WindowTitleHint)

        self.status = self.CANCEL

        self.title_label.setText(self.title_label.text() + igame.title)

        self.date_info_local.setText(dt_local.strftime("%A, %d. %B %Y %X"))
        self.date_info_remote.setText(dt_remote.strftime("%A, %d. %B %Y %X"))

        new_text = self.tr(" (newer)")
        if newer == "remote":
            self.cloud_gb.setTitle(self.cloud_gb.title() + new_text)
        elif newer == "local":
            self.local_gb.setTitle(self.local_gb.title() + new_text)

        self.icon_local.setPixmap(icon("mdi.harddisk", "fa.desktop").pixmap(128, 128))
        self.icon_remote.setPixmap(icon("mdi.cloud-outline", "ei.cloud").pixmap(128, 128))

        self.upload_button.clicked.connect(lambda: self.btn_clicked(self.UPLOAD))
        self.download_button.clicked.connect(lambda: self.btn_clicked(self.DOWNLOAD))
        self.cancel_button.clicked.connect(self.close)

        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
        self.layout().setSizeConstraint(QLayout.SetFixedSize)

    def get_action(self):
        self.exec_()
        return self.status

    def btn_clicked(self, status):
        self.status = status
        self.close()


class CloudSaveUtils(QObject):
    sync_finished = pyqtSignal(str)

    def __init__(self):
        super(CloudSaveUtils, self).__init__()
        self.core = LegendaryCoreSingleton()
        self.args = ArgumentsSingleton()
        self.api_results = ApiResultsSingleton()
        saves = self.api_results.saves
        if not self.args.offline:
            self.latest_saves = self.get_latest_saves(saves)
        else:
            self.latest_saves = dict()
        self.settings = QSettings()

        self.thread_pool = QThreadPool.globalInstance()

    def get_latest_saves(self, saves: List[SaveGameFile]) -> Dict[str, SaveGameFile]:
        save_games = set()
        for igame in self.core.get_installed_list():
            game = self.core.get_game(igame.app_name)
            if self.core.is_installed(igame.app_name) and game.supports_cloud_saves:
                save_games.add(igame.app_name)

        latest_saves = dict()
        for s in sorted(saves, key=lambda a: a.datetime):
            if s.app_name in save_games:
                latest_saves[s.app_name] = s
        return latest_saves

    def sync_before_launch_game(self, app_name, ignore_settings=False) -> int:
        if not ignore_settings:
            default = self.settings.value("auto_sync_cloud", True, bool)
            if not self.settings.value(f"{app_name}/auto_sync_cloud", default, bool):
                return False

        igame = self.core.get_installed_game(app_name)

        if not igame.save_path:
            try:
                savepath = self.core.get_save_path(app_name)
            except Exception as e:
                logger.error(e)
                savepath = ""
            if savepath:
                igame.save_path = savepath
                self.core.lgd.set_installed_game(app_name, igame)
                logger.info(f"Set save path of {igame.title} to {savepath}")
            elif not ignore_settings:  # sync on startup
                if (
                        QMessageBox.question(
                            None,
                            "Warning",
                            self.tr(
                                "Could not compute cloud save path. Please set it in Game settings manually. \nDo you want to launch {} anyway?"
                            ).format(igame.title),
                            QMessageBox.Yes | QMessageBox.No,
                            QMessageBox.No,
                        )
                        == QMessageBox.Yes
                ):
                    return False
                else:
                    raise ValueError("No savepath detected")
            else:
                QMessageBox.warning(
                    None,
                    "Warning",
                    self.tr(
                        "No savepath found. Please set it in Game Settings manually"
                    ),
                )
                return False

        res, (dt_local, dt_remote) = self.core.check_savegame_state(
            igame.save_path, self.latest_saves.get(app_name)
        )

        if res == SaveGameStatus.NO_SAVE:
            return False

        if res != SaveGameStatus.SAME_AGE:
            newer = ""
            if res == SaveGameStatus.REMOTE_NEWER:
                newer = "remote"
            elif res == SaveGameStatus.LOCAL_NEWER:
                newer = "local"

            if res == SaveGameStatus.REMOTE_NEWER and not dt_local:
                self.download_saves(igame)
                return True

            elif res == SaveGameStatus.LOCAL_NEWER and not dt_remote:
                self.upload_saves(igame, dt_local)
                return True

            result = CloudSaveDialog(igame, dt_local, dt_remote, newer).get_action()
            if result == CloudSaveDialog.UPLOAD:
                self.upload_saves(igame, dt_local)
            elif result == CloudSaveDialog.DOWNLOAD:
                self.download_saves(igame)
            elif result == CloudSaveDialog.CANCEL:
                raise AssertionError()
            return True

        return False

    def game_finished(self, app_name, ignore_settings=False, always_ask: bool = False):
        if not ignore_settings:
            default = self.settings.value("auto_sync_cloud", True, bool)
            if not self.settings.value(f"{app_name}/auto_sync_cloud", default, bool):
                self.sync_finished.emit(app_name)
                return

        igame = self.core.get_installed_game(app_name)

        if not igame.save_path:
            try:
                savepath = self.core.get_save_path(app_name)
            except Exception as e:
                logger.error(e)
                savepath = ""
            if savepath:
                igame.save_path = savepath
                self.core.lgd.set_installed_game(app_name, igame)
                logger.info(f"Set save path of {igame.title} to {savepath}")
            else:
                QMessageBox.warning(
                    None, "Warning", self.tr("No savepath set. Skip syncing with cloud")
                )
                return False

        res, (dt_local, dt_remote) = self.core.check_savegame_state(
            igame.save_path, self.latest_saves.get(app_name)
        )

        if res == SaveGameStatus.LOCAL_NEWER and not always_ask:
            self.upload_saves(igame, dt_local)
            return

        elif res == SaveGameStatus.NO_SAVE and not always_ask:
            QMessageBox.warning(
                None,
                "No saves",
                self.tr(
                    "There are no saves local and online. Maybe you have to change save path of {}"
                ).format(igame.title),
            )
            self.sync_finished.emit(app_name)
            return

        elif res == SaveGameStatus.SAME_AGE and not always_ask:
            self.sync_finished.emit(app_name)
            return

        # Remote newer
        newer = ""
        if res == SaveGameStatus.REMOTE_NEWER:
            newer = "remote"
        elif res == SaveGameStatus.LOCAL_NEWER:
            newer = "local"
        result = CloudSaveDialog(igame, dt_local, dt_remote, newer).get_action()
        if result == CloudSaveDialog.UPLOAD:
            self.upload_saves(igame, dt_local)
        elif result == CloudSaveDialog.DOWNLOAD:
            self.download_saves(igame)

    def upload_saves(self, igame: InstalledGame, dt_local):
        logger.info(f"Uploading saves for {igame.title}")
        w = SaveWorker(UploadModel(igame.app_name, dt_local, igame.save_path))
        w.signals.finished.connect(self.worker_finished)
        self.thread_pool.start(w)

    def download_saves(self, igame):
        logger.info(f"Downloading saves for {igame.title}")
        w = SaveWorker(
            DownloadModel(
                igame.app_name, self.latest_saves.get(igame.app_name), igame.save_path
            )
        )
        w.signals.finished.connect(self.worker_finished)
        self.thread_pool.start(w)

    def worker_finished(self, error_message: str, app_name: str):
        if not error_message:

            self.sync_finished.emit(app_name)
            self.latest_saves = self.get_latest_saves(self.api_results.saves)
        else:
            QMessageBox.warning(
                None,
                "Warning",
                self.tr("Syncing with cloud failed: \n {}").format(error_message)
            )
            self.sync_finished.emit(app_name)


def test_dialog():
    app = QApplication(sys.argv)
    core = LegendaryCore()
    dlg = CloudSaveDialog(core.get_installed_list()[0], datetime.datetime.now(),
                          datetime.datetime.strptime("2021,1", "%Y,%M"), "local")
    dlg.show()
    app.exec_()


if __name__ == '__main__':
    test_dialog()



from typing import Tuple

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QDialog,
    QLabel,
    QVBoxLayout,
    QCheckBox,
    QHBoxLayout,
    QPushButton,
)

from legendary.models.game import Game
from legendary.utils.selective_dl import get_sdl_appname

from rare.utils.misc import icon


class UninstallDialog(QDialog):
    def __init__(self, game: Game):
        super(UninstallDialog, self).__init__()
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setWindowTitle("Uninstall Game")
        layout = QVBoxLayout()
        self.info_text = QLabel(
            self.tr("Do you really want to uninstall <b>{}</b>?").format(game.app_title)
        )
        layout.addWidget(self.info_text)
        self.keep_files = QCheckBox(self.tr("Keep game files."))
        self.keep_config = QCheckBox(self.tr("Keep game configuation."))
        form_layout = QVBoxLayout()
        form_layout.setContentsMargins(6, 6, 0, 6)
        form_layout.addWidget(self.keep_files)
        form_layout.addWidget(self.keep_config)
        layout.addLayout(form_layout)

        button_layout = QHBoxLayout()
        self.ok_button = QPushButton(
            icon("ei.remove-circle", color="red"), self.tr("Uninstall")
        )
        self.ok_button.clicked.connect(self.ok)

        self.cancel_button = QPushButton(self.tr("Cancel"))
        self.cancel_button.clicked.connect(self.cancel)

        button_layout.addWidget(self.ok_button)
        button_layout.addStretch(1)
        button_layout.addWidget(self.cancel_button)
        layout.addLayout(button_layout)
        self.setLayout(layout)

        if get_sdl_appname(game.app_name) is not None:
            self.keep_config.setChecked(True)

        self.options: Tuple[bool, bool, bool] = (
            False, self.keep_files.isChecked(), self.keep_config.isChecked()
        )

    def get_options(self) -> Tuple[bool, bool, bool]:
        self.exec_()
        return self.options

    def ok(self):
        self.options = (True, self.keep_files.isChecked(), self.keep_config.isChecked())
        self.close()

    def cancel(self):
        self.options = (False, False, False)
        self.close()



from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFontMetrics, QPaintEvent
from PyQt5.QtWidgets import QLabel


class ElideLabel(QLabel):
    __text: str = ""

    def __init__(self, text="", parent=None, flags=Qt.WindowFlags()):
        super(ElideLabel, self).__init__(parent=parent, flags=flags)
        if text:
            self.setText(text)

    def setText(self, a0: str) -> None:
        self.__text = a0
        self.__setElideText(a0)

    def __setElideText(self, a0: str):
        metrics = QFontMetrics(self.font())
        elided_text = metrics.elidedText(a0, Qt.ElideRight, self.width())
        super(ElideLabel, self).setText(elided_text)

    def paintEvent(self, a0: QPaintEvent) -> None:
        self.__setElideText(self.__text)
        super(ElideLabel, self).paintEvent(a0)



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QWidget, QMessageBox
from legendary.models.game import Game

from rare.components.tabs.games.game_utils import GameUtils
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton, ImageManagerSingleton
from rare.shared.image_manager import ImageSize
from rare.ui.components.tabs.games.game_info.game_dlc import Ui_GameDlc
from rare.ui.components.tabs.games.game_info.game_dlc_widget import Ui_GameDlcWidget
from rare.models.install import InstallOptionsModel
from rare.widgets.image_widget import ImageWidget


class GameDlc(QWidget, Ui_GameDlc):
    install_dlc = pyqtSignal(str, bool)
    game: Game

    def __init__(self, dlcs: dict, game_utils: GameUtils, parent=None):
        super(GameDlc, self).__init__(parent=parent)
        self.setupUi(self)
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        self.game_utils = game_utils

        self.available_dlc_scroll.setFrameStyle(QFrame.NoFrame)
        self.installed_dlc_scroll.setFrameStyle(QFrame.NoFrame)

        self.dlcs = dlcs
        self.installed_dlc_widgets = list()
        self.available_dlc_widgets = list()

    def update_dlcs(self, app_name):
        self.game = self.core.get_game(app_name)
        dlcs = self.dlcs[self.game.catalog_item_id]
        self.title.setTitle(self.game.app_title)

        if self.installed_dlc_widgets:
            for dlc_widget in self.installed_dlc_widgets:
                dlc_widget.uninstall.disconnect()
                dlc_widget.deleteLater()
        self.installed_dlc_widgets.clear()
        if self.available_dlc_widgets:
            for dlc_widget in self.available_dlc_widgets:
                dlc_widget.install.disconnect()
                dlc_widget.deleteLater()
        self.available_dlc_widgets.clear()

        for dlc in sorted(dlcs, key=lambda x: x.app_title):
            if self.core.is_installed(dlc.app_name):
                dlc_widget = GameDlcWidget(dlc, True)
                self.installed_dlc_contents_layout.addWidget(dlc_widget)
                dlc_widget.uninstall.connect(self.uninstall)
                self.installed_dlc_widgets.append(dlc_widget)

            else:
                dlc_widget = GameDlcWidget(dlc, False)
                self.available_dlc_contents_layout.addWidget(dlc_widget)
                dlc_widget.install.connect(self.install)
                self.available_dlc_widgets.append(dlc_widget)

        self.installed_dlc_label.setVisible(not self.installed_dlc_widgets)
        self.installed_dlc_scroll.setVisible(bool(self.installed_dlc_widgets))

        self.available_dlc_label.setVisible(not self.available_dlc_widgets)
        self.available_dlc_scroll.setVisible(bool(self.available_dlc_widgets))

    def uninstall(self, app_name):
        if self.game_utils.uninstall_game(app_name):
            self.update_dlcs(self.game.app_name)

    def install(self, app_name):
        if not self.core.is_installed(self.game.app_name):
            QMessageBox.warning(
                self,
                "Error",
                self.tr("Base Game is not installed. Please install {} first").format(
                    self.game.app_title
                ),
            )
            return

        self.signals.install_game.emit(
            InstallOptionsModel(app_name=app_name, update=True)
        )


class GameDlcWidget(QFrame, Ui_GameDlcWidget):
    install = pyqtSignal(str)  # Appname
    uninstall = pyqtSignal(str)

    def __init__(self, dlc: Game, installed: bool, parent=None):
        super(GameDlcWidget, self).__init__(parent=parent)
        self.image_manager = ImageManagerSingleton()
        self.setupUi(self)
        self.dlc = dlc

        self.image = ImageWidget(self)
        self.image.setFixedSize(ImageSize.Smaller)
        self.dlc_layout.insertWidget(0, self.image)

        self.dlc_name.setText(dlc.app_title)
        self.version.setText(dlc.app_version())
        self.app_name.setText(dlc.app_name)

        self.image.setPixmap(self.image_manager.get_pixmap(dlc.app_name))

        if installed:
            self.action_button.setProperty("uninstall", 1)
            self.action_button.clicked.connect(self.uninstall_dlc)
            self.action_button.setText(self.tr("Uninstall DLC"))
        else:
            self.action_button.setProperty("install", 1)
            self.action_button.clicked.connect(self.install_game)
            self.action_button.setText(self.tr("Install DLC"))

    def uninstall_dlc(self):
        self.action_button.setDisabled(True)
        self.action_button.setText(self.tr("Uninstalling"))
        self.uninstall.emit(self.dlc.app_name)

    def install_game(self):
        self.action_button.setDisabled(True)
        self.action_button.setText(self.tr("Installing"))
        self.install.emit(self.dlc.app_name)



import os
import platform
from logging import getLogger

from PyQt5.QtCore import pyqtSignal, QProcess, QSettings, QStandardPaths, Qt, QByteArray
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import QFrame, QMessageBox, QAction

from rare.components.tabs.games.game_utils import GameUtils
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton, ArgumentsSingleton, ImageManagerSingleton
from rare.shared.image_manager import ImageSize
from rare.utils.misc import create_desktop_link
from rare.widgets.image_widget import ImageWidget

logger = getLogger("Game")


class BaseInstalledWidget(QFrame):
    launch_signal = pyqtSignal(str, QProcess, list)
    show_info = pyqtSignal(str)
    finish_signal = pyqtSignal(str, int)
    proc: QProcess()

    def __init__(self, app_name, pixmap: QPixmap, game_utils: GameUtils):
        super(BaseInstalledWidget, self).__init__()
        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()
        self.args = ArgumentsSingleton()
        self.image_manager = ImageManagerSingleton()
        self.game_utils = game_utils

        self.syncing_cloud_saves = False

        self.texts = {
            "needs_verification": self.tr("Please verify game before playing"),
            "hover": {
                "update_available": self.tr("Start game without version check"),
                "launch": self.tr("Launch Game"),
                "launch_origin": self.tr("Launch/Link"),
                "running": self.tr("Game running"),
                "launch_offline": self.tr("Launch offline"),
                "no_launch": self.tr("Can't launch game")
            },
            "default": {
                "running": self.tr("Game running"),
                "syncing": self.tr("Syncing cloud saves"),
                "update_available": self.tr("Update available"),
                "no_meta": self.tr("Game is only offline available")
            },
        }

        self.game = self.core.get_game(app_name)
        self.igame = self.core.get_installed_game(app_name)  # None if origin

        if self.game.app_title == "Unreal Engine":
            self.game.app_title = f"{self.game.app_title} {self.game.app_name.split('_')[-1]}"

        self.is_only_offline = False

        try:
            self.core.get_asset(app_name, platform=self.igame.platform).build_version
        except ValueError:
            logger.warning(f"Game {self.game.app_title} has no metadata. Set offline true")
            self.is_only_offline = True
        except AttributeError:
            pass

        self.image = ImageWidget(self)
        self.image.setFixedSize(ImageSize.Display)
        self.image.setPixmap(pixmap)
        self.game_running = False
        self.offline = self.args.offline
        self.update_available = False
        if self.igame and self.core.lgd.assets:
            try:
                remote_version = self.core.get_asset(
                    self.game.app_name, platform=self.igame.platform, update=False
                ).build_version
            except ValueError:
                logger.error(f"Asset error for {self.game.app_title}")
                self.update_available = False
            else:
                if remote_version != self.igame.version:
                    self.update_available = True

        self.data = QByteArray()
        self.settings = QSettings()

        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        launch = QAction(self.tr("Launch"), self)
        launch.triggered.connect(self.launch)
        self.addAction(launch)

        if self.game.supports_cloud_saves:
            sync = QAction(self.tr("Sync with cloud"), self)
            sync.triggered.connect(self.sync_game)
            self.addAction(sync)

        desktop = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
        if os.path.exists(
                os.path.join(desktop, f"{self.game.app_title}.desktop")
        ) or os.path.exists(os.path.join(desktop, f"{self.game.app_title}.lnk")):
            self.create_desktop = QAction(self.tr("Remove Desktop link"))
        else:
            self.create_desktop = QAction(self.tr("Create Desktop link"))
        if self.igame:
            self.create_desktop.triggered.connect(
                lambda: self.create_desktop_link("desktop")
            )
            self.addAction(self.create_desktop)

        applications = QStandardPaths.writableLocation(QStandardPaths.ApplicationsLocation)
        if platform.system() == "Linux":
            start_menu_file = os.path.join(applications, f"{self.game.app_title}.desktop")
        elif platform.system() == "Windows":
            start_menu_file = os.path.join(applications, "..", f"{self.game.app_title}.lnk")
        else:
            start_menu_file = ""
        if platform.system() in ["Windows", "Linux"]:
            if os.path.exists(start_menu_file):
                self.create_start_menu = QAction(self.tr("Remove start menu link"))
            else:
                self.create_start_menu = QAction(self.tr("Create start menu link"))
            if self.igame:
                self.create_start_menu.triggered.connect(
                    lambda: self.create_desktop_link("start_menu")
                )
                self.addAction(self.create_start_menu)

        reload_image = QAction(self.tr("Reload Image"), self)
        reload_image.triggered.connect(self.reload_image)
        self.addAction(reload_image)

        if self.igame is not None:
            uninstall = QAction(self.tr("Uninstall"), self)
            self.addAction(uninstall)
            uninstall.triggered.connect(
                lambda: self.signals.update_gamelist.emit([self.game.app_name])
                if self.game_utils.uninstall_game(self.game.app_name)
                else None
            )

    def reload_image(self):
        self.image_manager.download_image_blocking(self.game, force=True)
        pm = self.image_manager.get_pixmap(self.game.app_name, color=True)
        self.image.setPixmap(pm)

    def create_desktop_link(self, type_of_link):
        if type_of_link == "desktop":
            shortcut_path = QStandardPaths.writableLocation(QStandardPaths.DesktopLocation)
        elif type_of_link == "start_menu":
            shortcut_path = QStandardPaths.writableLocation(QStandardPaths.ApplicationsLocation)
        else:
            return

        if platform.system() == "Windows":
            shortcut_path = os.path.join(shortcut_path, f"{self.game.app_title}.lnk")
        elif platform.system() == "Linux":
            shortcut_path = os.path.join(shortcut_path, f"{self.game.app_title}.desktop")
        else:
            QMessageBox.warning(
                self,
                "Warning",
                f"Create a Desktop link is currently not supported on {platform.system()}",
            )
            return

        if not os.path.exists(shortcut_path):
            try:
                if not create_desktop_link(self.game.app_name, self.core, type_of_link):
                    return
            except PermissionError:
                QMessageBox.warning(
                    self, "Error", "Permission error. Cannot create Desktop Link"
                )
            if type_of_link == "desktop":
                self.create_desktop.setText(self.tr("Remove Desktop link"))
            elif type_of_link == "start_menu":
                self.create_start_menu.setText(self.tr("Remove Start menu link"))
        else:
            if os.path.exists(shortcut_path):
                os.remove(shortcut_path)

            if type_of_link == "desktop":
                self.create_desktop.setText(self.tr("Create Desktop link"))
            elif type_of_link == "start_menu":
                self.create_start_menu.setText(self.tr("Create Start menu link"))

    def launch(self, offline=False, skip_version_check=False):
        if self.game_running:
            return
        offline = offline or self.is_only_offline
        if self.is_only_offline and not self.igame.can_run_offline:
            QMessageBox.warning(self, "Warning",
                                self.tr("This game is probably not in your library and it cannot be launched offline"))
            return

        if self.game.supports_cloud_saves and not offline:
            self.syncing_cloud_saves = True
        self.game_utils.prepare_launch(
            self.game.app_name, offline, skip_version_check
        )

    def sync_finished(self, app_name):
        self.syncing_cloud_saves = False

    def sync_game(self):
        try:
            sync = self.game_utils.cloud_save_utils.sync_before_launch_game(
                self.game.app_name, True
            )
        except Exception:
            sync = False
        if sync:
            self.syncing_cloud_saves = True

    def game_finished(self, app_name, error):
        if error:
            QMessageBox.warning(self, "Error", error)
        self.game_running = False

    # From RareGame, added from sorting to work
    @property
    def is_non_asset(self) -> bool:
        """!
        @brief Property to report if a Game doesn't have assets

        Typically, games have assets, however some games that require
        other launchers do not have them. Rare treats these games as installed
        offering to execute their launcher.

        @return bool If the game doesn't have assets
        """
        return not self.game.asset_infos

    @property
    def is_origin(self) -> bool:
        return self.game.metadata.get("customAttributes", {}).get("ThirdPartyManagedApp", {}).get("value") == "Origin"

    @property
    def is_installed(self) -> bool:
        """!
        @brief Property to report if a game is installed

        This returns True if InstalledGame data have been loaded for the game
        or if the game is a game without assets, for example an Origin game.

        @return bool If the game should be considered installed
        """
        return (self.igame is not None) or self.is_non_asset

    @property
    def app_title(self) -> str:
        return self.igame.title if self.igame is not None else self.game.app_title



import os
import platform as pf
from dataclasses import dataclass
from typing import List, Optional, Callable, Dict

from legendary.models.downloading import AnalysisResult, ConditionCheckResult
from legendary.models.game import Game, InstalledGame

from rare.lgndr.downloader.mp.manager import DLManager


@dataclass
class InstallOptionsModel:
    app_name: str
    base_path: str = ""
    shared_memory: int = 1024
    max_workers: int = os.cpu_count() * 2
    force: bool = False
    platform: str = "Windows"
    install_tag: Optional[List[str]] = None
    order_opt: bool = False
    repair_mode: bool = False
    repair_and_update: bool = False
    no_install: bool = False
    ignore_space: bool = False
    skip_dlcs: bool = False
    with_dlcs: bool = False
    # Rare's internal arguments
    # FIXME: Do we really need all of these?
    create_shortcut: bool = True
    overlay: bool = False
    update: bool = False
    silent: bool = False
    install_prereqs: bool = pf.system() == "Windows"

    def __post_init__(self):
        self.sdl_prompt: Callable[[str, str], list] = \
            lambda app_name, title: self.install_tag if self.install_tag is not None else [""]

    def as_install_kwargs(self) -> Dict:
        return {
            k: getattr(self, k)
            for k in self.__dict__
            if k not in ["update", "silent", "create_shortcut", "overlay", "install_prereqs"]
        }


@dataclass
class InstallDownloadModel:
    dlm: DLManager
    analysis: AnalysisResult
    igame: InstalledGame
    game: Game
    repair: bool
    repair_file: str
    res: ConditionCheckResult


@dataclass
class InstallQueueItemModel:
    download: Optional[InstallDownloadModel] = None
    options: Optional[InstallOptionsModel] = None

    def __bool__(self):
        return (self.download is not None) and (self.options is not None)



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QPushButton

from rare.shared import GlobalSignalsSingleton


class DebugSettings(QWidget):
    def __init__(self):
        super(DebugSettings, self).__init__()
        self.setLayout(QVBoxLayout())

        self.raise_runtime_exception_button = QPushButton("Raise Exception")
        self.layout().addWidget(self.raise_runtime_exception_button)
        self.raise_runtime_exception_button.clicked.connect(self.raise_exception)
        self.restart_button = QPushButton("Restart")
        self.layout().addWidget(self.restart_button)
        self.restart_button.clicked.connect(lambda: GlobalSignalsSingleton().exit_app.emit(-133742))

        self.layout().addStretch(1)

    def raise_exception(self):
        raise RuntimeError("Debug Crash")



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'rare/ui/components/tabs/settings/widgets/env_vars.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_EnvVars(object):
    def setupUi(self, EnvVars):
        EnvVars.setObjectName("EnvVars")
        EnvVars.resize(720, 561)
        EnvVars.setMinimumSize(QtCore.QSize(0, 200))
        EnvVars.setWindowTitle("GroupBox")
        self.verticalLayout = QtWidgets.QVBoxLayout(EnvVars)
        self.verticalLayout.setObjectName("verticalLayout")
        self.env_vars_table = QtWidgets.QTableWidget(EnvVars)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.env_vars_table.sizePolicy().hasHeightForWidth())
        self.env_vars_table.setSizePolicy(sizePolicy)
        self.env_vars_table.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAsNeeded)
        self.env_vars_table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)
        self.env_vars_table.setObjectName("env_vars_table")
        self.env_vars_table.setColumnCount(2)
        self.env_vars_table.setRowCount(1)
        item = QtWidgets.QTableWidgetItem()
        item.setText("-")
        self.env_vars_table.setVerticalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.env_vars_table.setHorizontalHeaderItem(0, item)
        item = QtWidgets.QTableWidgetItem()
        self.env_vars_table.setHorizontalHeaderItem(1, item)
        self.env_vars_table.horizontalHeader().setStretchLastSection(True)
        self.env_vars_table.verticalHeader().setStretchLastSection(False)
        self.verticalLayout.addWidget(self.env_vars_table)

        self.retranslateUi(EnvVars)
        QtCore.QMetaObject.connectSlotsByName(EnvVars)

    def retranslateUi(self, EnvVars):
        _translate = QtCore.QCoreApplication.translate
        EnvVars.setTitle(_translate("EnvVars", "Environment variables"))
        item = self.env_vars_table.horizontalHeaderItem(0)
        item.setText(_translate("EnvVars", "Key"))
        item = self.env_vars_table.horizontalHeaderItem(1)
        item.setText(_translate("EnvVars", "Value"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    EnvVars = QtWidgets.QGroupBox()
    ui = Ui_EnvVars()
    ui.setupUi(EnvVars)
    EnvVars.show()
    sys.exit(app.exec_())



import webbrowser
from logging import getLogger

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QWidget

from rare import __version__, code_name
from rare.ui.components.tabs.settings.about import Ui_About
from rare.utils.qt_requests import QtRequestManager

logger = getLogger("About")


def versiontuple(v):
    try:
        return tuple(map(int, (v.split("."))))
    except:
        return tuple((9, 9, 9))  # It is a beta version and newer


class About(QWidget, Ui_About):
    update_available = False
    update_available_ready = pyqtSignal()

    def __init__(self):
        super(About, self).__init__()
        self.setupUi(self)

        self.version.setText(f"{__version__}  {code_name}")

        self.update_label.setVisible(False)
        self.update_lbl.setVisible(False)
        self.open_browser.setVisible(False)

        self.manager = QtRequestManager("json")
        self.manager.get(
            "https://api.github.com/repos/Dummerle/Rare/releases/latest",
            self.update_available_finished,
        )

        self.open_browser.clicked.connect(
            lambda: webbrowser.open("https://github.com/Dummerle/Rare/releases/latest")
        )

    def update_available_finished(self, data: dict):
        if latest_tag := data.get("tag_name"):
            self.update_available = versiontuple(latest_tag) > versiontuple(__version__)
        else:
            self.update_available = False

        if self.update_available:
            logger.info(f"Update available: {__version__} -> {latest_tag}")
            self.update_lbl.setText(
                self.tr("Update available: {} -> {}").format(__version__, latest_tag)
            )
            self.update_label.setVisible(True)
            self.update_lbl.setVisible(True)
            self.open_browser.setVisible(True)
            self.update_available_ready.emit()



import os
import shutil
from logging import getLogger

from PyQt5.QtWidgets import QFileDialog, QWidget

from rare.components.tabs.settings.widgets.dxvk import DxvkSettings
from rare.components.tabs.settings.widgets.mangohud import MangoHudSettings
from rare.shared import LegendaryCoreSingleton, GlobalSignalsSingleton

from rare.ui.components.tabs.settings.linux import Ui_LinuxSettings
from rare.utils.extra_widgets import PathEdit

logger = getLogger("LinuxSettings")


class LinuxSettings(QWidget, Ui_LinuxSettings):
    def __init__(self, name=None):
        super(LinuxSettings, self).__init__()
        self.setupUi(self)

        self.core = LegendaryCoreSingleton()
        self.signals = GlobalSignalsSingleton()

        self.name = name if name is not None else "default"

        # Wine prefix
        self.wine_prefix = PathEdit(
            self.load_prefix(),
            file_type=QFileDialog.DirectoryOnly,
            edit_func=lambda path: (os.path.isdir(path) or not path, path, PathEdit.reasons.dir_not_exist),
            save_func=self.save_prefix,
        )
        self.prefix_layout.addWidget(self.wine_prefix)

        # Wine executable
        self.wine_exec = PathEdit(
            self.load_setting(self.name, "wine_executable"),
            file_type=QFileDialog.ExistingFile,
            name_filter="Wine executable (wine wine64)",
            edit_func=lambda text: (os.path.exists(text) or not text, text, PathEdit.reasons.dir_not_exist),
            save_func=lambda text: self.save_setting(
                text, section=self.name, setting="wine_executable"
            ),
        )
        self.exec_layout.addWidget(self.wine_exec)

        # dxvk
        self.dxvk = DxvkSettings()
        self.overlay_layout.addWidget(self.dxvk)
        self.dxvk.load_settings(self.name)

        self.mangohud = MangoHudSettings()
        self.overlay_layout.addWidget(self.mangohud)
        self.mangohud.load_settings(self.name)

        if not shutil.which("mangohud"):
            self.mangohud.setDisabled(True)
            self.mangohud.setToolTip(self.tr("Mangohud is not installed or not in path"))

    def load_prefix(self) -> str:
        return self.load_setting(
            f"{self.name}.env",
            "WINEPREFIX",
            fallback=self.load_setting(self.name, "wine_prefix"),
        )

    def save_prefix(self, text: str):
        self.save_setting(text, f"{self.name}.env", "WINEPREFIX")
        self.save_setting(text, self.name, "wine_prefix")
        self.signals.wine_prefix_updated.emit()

    def load_setting(self, section: str, setting: str, fallback: str = str()):
        return self.core.lgd.config.get(section, setting, fallback=fallback)

    def save_setting(self, text: str, section: str, setting: str):
        if text:
            if section not in self.core.lgd.config.sections():
                self.core.lgd.config.add_section(section)
                logger.debug(f"Added {f'[{section}]'} configuration section")
            self.core.lgd.config.set(section, setting, text)
            logger.debug(f"Set {setting} in {f'[{section}]'} to {text}")

        else:
            if self.core.lgd.config.has_section(
                section
            ) and self.core.lgd.config.has_option(section, setting):
                self.core.lgd.config.remove_option(section, setting)
                logger.debug(f"Unset {setting} from {f'[{section}]'}")
                if not self.core.lgd.config[section]:
                    self.core.lgd.config.remove_section(section)
                    logger.debug(f"Removed {f'[{section}]'} configuration section")

        self.core.lgd.save_config()



from logging import getLogger

from PyQt5.QtCore import QProcess, pyqtSignal, Qt
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QPushButton, QVBoxLayout

from rare.components.tabs.games.game_widgets.base_installed_widget import (
    BaseInstalledWidget,
)
from rare.utils.misc import icon, get_size

logger = getLogger("GameWidget")


class InstalledListWidget(BaseInstalledWidget):
    proc: QProcess
    signal = pyqtSignal(str)
    update_game = pyqtSignal()

    def __init__(self, app_name, pixmap, game_utils):
        super(InstalledListWidget, self).__init__(app_name, pixmap, game_utils)
        self.setFrameStyle(self.StyledPanel)
        self.dev = self.game.metadata["developer"]
        if self.igame:
            self.size = self.igame.install_size
        else:
            self.size = 0

        self.layout = QHBoxLayout()
        self.setLayout(self.layout)
        self.layout.addWidget(self.image)

        # Layout on the right
        self.childLayout = QVBoxLayout()
        self.layout.addLayout(self.childLayout)

        play_icon = icon("ei.play")
        self.title_label = QLabel(f"<h1>{self.game.app_title}</h1>")
        self.title_label.setWordWrap(True)
        self.childLayout.addWidget(self.title_label)
        self.app_name_label = QLabel(self.game.app_name)
        self.launch_button = QPushButton(
            play_icon,
            self.tr("Launch") if self.igame else self.tr("Link/Play") if self.is_origin else self.texts["hover"][
                "no_launch"]
        )
        if not self.is_origin and not self.igame:
            self.launch_button.setDisabled(True)

        self.launch_button.setObjectName("launch_game_button")
        self.launch_button.setFixedWidth(150)

        self.info = QPushButton("Info")
        self.info.clicked.connect(lambda: self.show_info.emit(self.game.app_name))
        self.info.setFixedWidth(80)

        self.info_label = QLabel("")
        self.childLayout.addWidget(self.info_label)
        self.update_text()

        self.launch_button.clicked.connect(self.launch)

        self.childLayout.addWidget(self.launch_button)
        self.childLayout.addWidget(self.info)
        self.childLayout.addWidget(self.app_name_label)
        self.developer_label = QLabel(self.tr("Developer: {}").format(self.dev))
        self.childLayout.addWidget(self.developer_label)
        if self.igame:
            self.version_label = QLabel(f"Version: {self.igame.version}")
            self.size_label = QLabel(
                f"{self.tr('Installed size')}: {get_size(self.size)}"
            )

            self.childLayout.addWidget(self.version_label)
            self.childLayout.addWidget(self.size_label)

        self.childLayout.setAlignment(Qt.AlignTop)
        self.layout.setAlignment(Qt.AlignLeft)

        self.game_utils.cloud_save_finished.connect(self.sync_finished)
        self.game_utils.finished.connect(self.game_finished)

        self.leaveEvent = self.update_text
        self.enterEvent = self.update_text

        self.game_utils.game_launched.connect(self.game_started)

    def update_text(self, e=None):
        if self.update_available:
            self.info_label.setText(self.texts["default"]["update_available"])
        elif self.is_only_offline:
            self.info_label.setText(self.texts["default"]["no_meta"])
        elif self.igame and self.igame.needs_verification:
            self.info_label.setText(self.texts["needs_verification"])
        elif self.syncing_cloud_saves:
            self.info_label.setText(self.texts["default"]["syncing"])
        else:
            self.info_label.setText("")

    def game_started(self, app_name):
        if app_name == self.game.app_name:
            self.game_running = True
            self.update_text()
            self.launch_button.setDisabled(True)

    def game_finished(self, app_name, error):
        if app_name != self.game.app_name:
            return
        super().game_finished(app_name, error)
        self.update_text(None)
        self.launch_button.setDisabled(False)



import os
import platform
from logging import getLogger

from PyQt5.QtCore import Qt, pyqtSignal, QRunnable, QObject, QThreadPool, QSettings, pyqtSlot
from PyQt5.QtWidgets import QDialog, QApplication
from legendary.models.game import Game
from requests.exceptions import ConnectionError, HTTPError

from rare.components.dialogs.login import LoginDialog
from rare.models.apiresults import ApiResults
from rare.shared import LegendaryCoreSingleton, ArgumentsSingleton, ApiResultsSingleton, ImageManagerSingleton
from rare.ui.components.dialogs.launch_dialog import Ui_LaunchDialog
from rare.utils import legendary_utils
from rare.utils.misc import CloudWorker
from rare.widgets.elide_label import ElideLabel

logger = getLogger("LaunchDialog")


class LaunchWorker(QRunnable):
    class Signals(QObject):
        progress = pyqtSignal(int, str)
        result = pyqtSignal(object, str)
        finished = pyqtSignal()

    def __init__(self):
        super(LaunchWorker, self).__init__()
        self.setAutoDelete(True)
        self.signals = LaunchWorker.Signals()
        self.core = LegendaryCoreSingleton()

    def run_real(self):
        pass

    def run(self):
        self.run_real()
        self.signals.deleteLater()


class ImageWorker(LaunchWorker):
    # FIXME: this is a middle-ground solution for concurrent downloads
    class DownloadSlot(QObject):
        def __init__(self, signals: LaunchWorker.Signals):
            super(ImageWorker.DownloadSlot, self).__init__()
            self.signals = signals
            self.counter = 0
            self.length = 0

        @pyqtSlot(object)
        def counter_inc(self, game: Game):
            self.signals.progress.emit(
                int(self.counter / self.length * 75),
                self.tr("Downloading image for <b>{}</b>").format(game.app_title)
            )
            self.counter += 1

    def __init__(self):
        super(ImageWorker, self).__init__()
        # FIXME: this is a middle-ground solution for concurrent downloads
        self.dl_slot = ImageWorker.DownloadSlot(self.signals)
        self.image_manager = ImageManagerSingleton()

    def tr(self, t) -> str:
        return QApplication.instance().translate(self.__class__.__name__, t)

    def run_real(self):
        # Download Images
        games, dlcs = self.core.get_game_and_dlc_list(update_assets=True, skip_ue=False)
        self.signals.result.emit((games, dlcs), "gamelist")
        dlc_list = [dlc[0] for dlc in dlcs.values()]

        na_games, na_dlcs = self.core.get_non_asset_library_items(force_refresh=False, skip_ue=False)
        self.signals.result.emit(na_games, "no_assets")
        na_dlc_list = [dlc[0] for dlc in na_dlcs.values()]

        game_list = games + dlc_list + na_games + na_dlc_list

        self.dl_slot.length = len(game_list)
        for i, game in enumerate(game_list):
            if game.app_title == "Unreal Engine":
                game.app_title += f" {game.app_name.split('_')[-1]}"
                self.core.lgd.set_game_meta(game.app_name, game)
            # self.image_manager.download_image_blocking(game)
            self.image_manager.download_image(game, self.dl_slot.counter_inc, priority=0)
        # FIXME: this is a middle-ground solution for concurrent downloads
        while self.dl_slot.counter < len(game_list):
            QApplication.instance().processEvents()
        self.dl_slot.deleteLater()

        igame_list = self.core.get_installed_list(include_dlc=True)

        # FIXME: incorporate installed game status checking here for now, still slow
        for i, igame in enumerate(igame_list):
            self.signals.progress.emit(
                int(i / len(igame_list) * 25) + 75,
                self.tr("Validating install for <b>{}</b>").format(igame.title)
            )
            if not os.path.exists(igame.install_path):
                # lk; since install_path is lost anyway, set keep_files to True
                # lk: to avoid spamming the log with "file not found" errors
                legendary_utils.uninstall_game(self.core, igame.app_name, keep_files=True)
                logger.info(f"Uninstalled {igame.title}, because no game files exist")
                continue
            # lk: games that don't have an override and can't find their executable due to case sensitivity
            # lk: will still erroneously require verification. This might need to be removed completely
            # lk: or be decoupled from the verification requirement
            if override_exe := self.core.lgd.config.get(igame.app_name, "override_exe", fallback=""):
                igame_executable = override_exe
            else:
                igame_executable = igame.executable
            if not os.path.exists(os.path.join(igame.install_path, igame_executable.replace("\\", "/").lstrip("/"))):
                igame.needs_verification = True
                self.core.lgd.set_installed_game(igame.app_name, igame)
                logger.info(f"{igame.title} needs verification")
            # FIXME: end

        self.signals.progress.emit(100, self.tr("Launching Rare"))
        self.signals.finished.emit()


class ApiRequestWorker(LaunchWorker):
    def __init__(self):
        super(ApiRequestWorker, self).__init__()
        self.settings = QSettings()

    def run_real(self) -> None:
        if self.settings.value("mac_meta", platform.system() == "Darwin", bool):
            try:
                result = self.core.get_game_and_dlc_list(update_assets=False, platform="Mac")
            except HTTPError:
                result = [], {}
        else:
            result = [], {}
        self.signals.result.emit(result, "mac")

        if self.settings.value("win32_meta", False, bool):
            try:
                result = self.core.get_game_and_dlc_list(update_assets=False, platform="Win32")
            except HTTPError:
                result = [], {}
        else:
            result = [], {}
        self.signals.result.emit(result, "32bit")


class LaunchDialog(QDialog):
    quit_app = pyqtSignal(int)
    start_app = pyqtSignal()
    completed = 0

    def __init__(self, parent=None):
        super(LaunchDialog, self).__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setWindowFlags(
            Qt.Window
            | Qt.Dialog
            | Qt.CustomizeWindowHint
            | Qt.WindowSystemMenuHint
            | Qt.WindowTitleHint
            | Qt.WindowMinimizeButtonHint
            | Qt.MSWindowsFixedSizeDialogHint
        )
        self.setWindowModality(Qt.WindowModal)
        self.ui = Ui_LaunchDialog()
        self.ui.setupUi(self)

        self.progress_info = ElideLabel(parent=self)
        self.layout().addWidget(self.progress_info)

        self.core = LegendaryCoreSingleton()
        self.args = ArgumentsSingleton()
        self.thread_pool = QThreadPool().globalInstance()
        self.api_results = ApiResults()

        self.login_dialog = LoginDialog(core=self.core, parent=parent)

    def login(self):
        do_launch = True
        try:
            if self.args.offline:
                pass
            else:
                QApplication.instance().processEvents()
                # Force an update check and notice in case there are API changes
                self.core.check_for_updates(force=True)
                self.core.force_show_update = True
                if self.core.login():
                    logger.info("You are logged in")
                else:
                    raise ValueError("You are not logged in. Open Login Window")
        except ValueError as e:
            logger.info(str(e))
            # Do not set parent, because it won't show a task bar icon
            # Update: Inherit the same parent as LaunchDialog
            do_launch = self.login_dialog.login()
        except (HTTPError, ConnectionError) as e:
            logger.warning(e)
            self.args.offline = True
        finally:
            if do_launch:
                if not self.args.silent:
                    self.show()
                self.launch()
            else:
                self.quit_app.emit(0)

    def start_api_requests(self):
        # gamelist and no_asset games are from Image worker
        api_worker = ApiRequestWorker()
        api_worker.signals.result.connect(self.handle_api_worker_result)
        self.thread_pool.start(api_worker)

    def launch(self):
        self.progress_info.setText(self.tr("Preparing Rare"))

        if not self.args.offline:
            image_worker = ImageWorker()
            image_worker.signals.result.connect(self.handle_api_worker_result)
            image_worker.signals.progress.connect(self.update_progress)
            # lk: start the api requests worker after the manifests have been downloaded
            # lk: to avoid force updating the assets twice and causing inconsistencies
            image_worker.signals.finished.connect(self.start_api_requests)
            image_worker.signals.finished.connect(self.finish)
            self.thread_pool.start(image_worker)

            # cloud save from another worker, because it is used in cloud_save_utils too
            cloud_worker = CloudWorker()
            cloud_worker.signals.result_ready.connect(lambda x: self.handle_api_worker_result(x, "saves"))
            self.thread_pool.start(cloud_worker)

        else:
            self.completed = 2
            if self.core.lgd.assets:
                (
                    self.api_results.game_list,
                    self.api_results.dlcs,
                ) = self.core.get_game_and_dlc_list(False)
                self.api_results.bit32_games = list(
                    map(lambda i: i.app_name, self.core.get_game_list(False, "Win32"))
                )
                self.api_results.mac_games = list(
                    map(lambda i: i.app_name, self.core.get_game_list(False, "Mac"))
                )
            else:
                logger.warning("No assets found. Falling back to empty game lists")
                self.api_results.game_list, self.api_results.dlcs = [], {}
                self.api_results.mac_games = self.api_results.bit32_games = []
            self.finish()

    def handle_api_worker_result(self, result, text):
        logger.debug(f"Api Request got from {text}")
        if text == "gamelist":
            if result:
                self.api_results.game_list, self.api_results.dlcs = result
            else:
                (
                    self.api_results.game_list,
                    self.api_results.dlcs,
                ) = self.core.get_game_and_dlc_list(False)
        elif text == "no_assets":
            self.api_results.no_asset_games = result if result else []

        elif text == "32bit":
            self.api_results.bit32_games = [i.app_name for i in result[0]] if result else []
        elif text == "mac":
            self.api_results.mac_games = [i.app_name for i in result[0]] if result else []

        elif text == "saves":
            self.api_results.saves = result

        if self.api_results:
            self.finish()

    @pyqtSlot(int, str)
    def update_progress(self, i: int, m: str):
        self.ui.progress_bar.setValue(i)
        self.progress_info.setText(m)

    def finish(self):
        self.completed += 1
        if self.completed >= 2:
            logger.info("App starting")
            ApiResultsSingleton(self.api_results)
            self.completed += 1
            self.start_app.emit()

    def reject(self) -> None:
        if self.completed >= 3:
            super(LaunchDialog, self).reject()
        else:
            pass



from abc import abstractmethod
from typing import Optional

from PyQt5.QtCore import QParallelAnimationGroup, Qt, QPropertyAnimation, QAbstractAnimation, QSize
from PyQt5.QtWidgets import QWidget, QFrame, QToolButton, QGridLayout, QSizePolicy, QGroupBox, QLabel

from rare.utils.misc import icon


class CollapsibleBase(object):
    """
    References:
        # Adapted from c++ version
        https://stackoverflow.com/questions/32476006/how-to-make-an-expandable-collapsable-section-widget-in-qt
        # Adapted from python version
        https://newbedev.com/how-to-make-an-expandable-collapsable-section-widget-in-qt
    """
    def __init__(self):
        self.animation_duration = None
        self.toggle_animation = None
        self.content_area = None
        self.content_toggle_animation = None

    def setup(self, animation_duration: int = 200):
        self.animation_duration = animation_duration
        self.content_area: Optional[QWidget] = None
        self.content_toggle_animation: Optional[QPropertyAnimation] = None

        # let the entire widget grow and shrink with its content
        self.toggle_animation = QParallelAnimationGroup(self)
        self.toggle_animation.addAnimation(QPropertyAnimation(self, b"minimumHeight"))
        self.toggle_animation.addAnimation(QPropertyAnimation(self, b"maximumHeight"))

    @abstractmethod
    def isChecked(self) -> bool:
        ...

    @abstractmethod
    def click(self) -> None:
        ...

    @abstractmethod
    def addToLayout(self, widget: QWidget) -> None:
        ...

    @abstractmethod
    def sizeHint(self) -> QSize:
        ...

    def animationStart(self, checked):
        direction = QAbstractAnimation.Forward if checked else QAbstractAnimation.Backward
        self.toggle_animation.setDirection(direction)
        self.toggle_animation.start()

    def setWidget(self, widget: QWidget):
        if widget is None or widget is self.content_area:
            return
        is_checked = self.isChecked()
        if self.content_area is not None:
            # Collapse the parent before replacing the child
            if is_checked:
                self.click()
            self.toggle_animation.removeAnimation(self.content_toggle_animation)
            self.content_area.setParent(None)
            self.content_area.deleteLater()

        self.content_area = widget
        self.content_area.setParent(self)
        self.content_area.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)

        # start out collapsed
        if not is_checked:
            self.content_area.setMaximumHeight(0)
            self.content_area.setMinimumHeight(0)

        self.content_toggle_animation = QPropertyAnimation(self.content_area, b"maximumHeight")
        self.toggle_animation.addAnimation(self.content_toggle_animation)
        self.addToLayout(self.content_area)
        collapsed_height = self.sizeHint().height()
        content_height = self.content_area.sizeHint().height()
        for i in range(self.toggle_animation.animationCount() - 1):
            spoiler_animation = self.toggle_animation.animationAt(i)
            spoiler_animation.setDuration(self.animation_duration)
            spoiler_animation.setStartValue(collapsed_height)
            spoiler_animation.setEndValue(collapsed_height + content_height)
        content_animation = self.toggle_animation.animationAt(self.toggle_animation.animationCount() - 1)
        content_animation.setDuration(self.animation_duration)
        content_animation.setStartValue(0)
        content_animation.setEndValue(content_height)
        if is_checked:
            self.click()


class CollapsibleFrame(QFrame, CollapsibleBase):
    def __init__(
        self, widget: QWidget = None, title: str = "", button_text: str = "", animation_duration: int = 200, parent=None
    ):
        super(CollapsibleFrame, self).__init__(parent=parent)
        self.setup(animation_duration)
        self.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)

        self.toggle_button = QToolButton(self)
        self.toggle_button.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self.toggle_button.setIcon(icon("fa.arrow-right"))
        self.toggle_button.setText(button_text)
        self.toggle_button.setCheckable(True)
        self.toggle_button.setChecked(False)

        self.title_label = QLabel(title)
        font = self.title_label.font()
        font.setBold(True)
        self.title_label.setFont(font)
        self.title_label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # don't waste space
        self.main_layout = QGridLayout(self)
        self.main_layout.setVerticalSpacing(0)
        self.main_layout.setContentsMargins(0, 0, 0, 0)
        self.main_layout.addWidget(self.toggle_button, 0, 0, 1, 1, Qt.AlignLeft)
        self.main_layout.addWidget(self.title_label, 0, 1, 1, 1)
        self.main_layout.setColumnStretch(1, 1)
        self.main_layout.setRowStretch(0, 0)
        self.main_layout.setRowStretch(1, 1)
        self.setLayout(self.main_layout)

        self.toggle_button.clicked.connect(self.animationStart)

        if widget is not None:
            self.setWidget(widget)

    def isChecked(self) -> bool:
        return self.toggle_button.isChecked()

    def click(self) -> None:
        self.toggle_button.click()

    def addToLayout(self, widget: QWidget) -> None:
        self.main_layout.addWidget(widget, 1, 0, 1, 2)

    def sizeHint(self) -> QSize:
        return super(CollapsibleFrame, self).sizeHint()

    def animationStart(self, checked):
        arrow_type = icon("fa.arrow-down") if checked else icon("fa.arrow-right")
        self.toggle_button.setIcon(arrow_type)
        super(CollapsibleFrame, self).animationStart(checked)


class CollapsibleGroupBox(QGroupBox, CollapsibleBase):
    def __init__(
        self, widget: QWidget = None, title: str = "", animation_duration: int = 200, parent=None
    ):
        super(CollapsibleGroupBox, self).__init__(parent=parent)
        self.setup(animation_duration)
        self.setTitle(title)
        self.setCheckable(True)
        self.setChecked(False)

        # don't waste space
        self.main_layout = QVBoxLayout(self)
        self.main_layout.setSpacing(0)
        self.main_layout.setContentsMargins(0, 0, 0, -1)
        self.setLayout(self.main_layout)

        self.toggled.connect(self.animationStart)

        if widget is not None:
            self.setWidget(widget)

    def isChecked(self) -> bool:
        return super(CollapsibleGroupBox, self).isChecked()

    def click(self) -> None:
        self.setChecked(not self.isChecked())

    def addToLayout(self, widget: QWidget) -> None:
        self.main_layout.addWidget(widget)

    def sizeHint(self) -> QSize:
        return super(CollapsibleGroupBox, self).sizeHint()


if __name__ == "__main__":
    import sys
    from PyQt5.QtWidgets import QApplication, QDialog, QVBoxLayout
    from rare.ui.components.dialogs.install_dialog_advanced import Ui_InstallDialogAdvanced
    from rare.utils.misc import set_style_sheet
    from rare.resources.stylesheets import RareStyle
    app = QApplication(sys.argv)

    set_style_sheet("RareStyle")

    ui_frame = Ui_InstallDialogAdvanced()
    widget_frame = QWidget()
    ui_frame.setupUi(widget_frame)
    collapsible_frame = CollapsibleFrame(widget_frame, title="Frame me!")
    collapsible_frame.setDisabled(False)

    def replace_func_frame(state):
        widget2_frame = QWidget()
        ui2_frame = Ui_InstallDialogAdvanced()
        ui2_frame.setupUi(widget2_frame)
        if state:
            ui2_frame.install_dialog_advanced_layout.removeRow(3)
            ui2_frame.install_dialog_advanced_layout.removeRow(4)
        collapsible_frame.setWidget(widget2_frame)

    ui_group = Ui_InstallDialogAdvanced()
    widget_group = QWidget()
    ui_group.setupUi(widget_group)
    collapsible_group = CollapsibleGroupBox(widget_group, title="Group me!")
    collapsible_group.setDisabled(False)

    def replace_func_group(state):
        widget2_group = QWidget()
        ui2_group = Ui_InstallDialogAdvanced()
        ui2_group.setupUi(widget2_group)
        if state:
            ui2_group.install_dialog_advanced_layout.removeRow(3)
            ui2_group.install_dialog_advanced_layout.removeRow(4)
        collapsible_group.setWidget(widget2_group)

    replace_button = QToolButton()
    replace_button.setText("Replace me!")
    replace_button.setCheckable(True)
    replace_button.setChecked(False)
    replace_button.clicked.connect(replace_func_frame)
    replace_button.clicked.connect(replace_func_group)

    dialog = QDialog()
    dialog.setLayout(QVBoxLayout())
    dialog.layout().addWidget(replace_button)
    dialog.layout().addWidget(collapsible_frame)
    dialog.layout().addWidget(collapsible_group)
    dialog.layout().setSizeConstraint(QVBoxLayout.SetFixedSize)
    dialog.show()
    sys.exit(app.exec_())




import configparser
import os
from argparse import Namespace
from logging import getLogger
from typing import Optional

from PyQt5.QtCore import QObject

from rare.lgndr.core import LegendaryCore
from rare.models.apiresults import ApiResults
from rare.models.signals import GlobalSignals
from .image_manager import ImageManager

logger = getLogger("RareCore")


class RareCore(QObject):
    _instance: Optional['RareCore'] = None

    def __init__(self, args: Namespace):
        if self._instance is not None:
            raise RuntimeError("RareCore already initialized")
        super(RareCore, self).__init__()
        self._args: Optional[Namespace] = None
        self._signals: Optional[GlobalSignals] = None
        self._core: Optional[LegendaryCore] = None
        self._image_manager: Optional[ImageManager] = None
        self._api_results: Optional[ApiResults] = None

        self.args(args)
        self.signals(init=True)
        self.core(init=True)
        self.image_manager(init=True)

        RareCore._instance = self

    @staticmethod
    def instance() -> 'RareCore':
        if RareCore._instance is None:
            raise RuntimeError("Uninitialized use of RareCore")
        return RareCore._instance

    def signals(self, init: bool = False) -> GlobalSignals:
        if self._signals is None and not init:
            raise RuntimeError("Uninitialized use of GlobalSignalsSingleton")
        if self._signals is not None and init:
            raise RuntimeError("GlobalSignals already initialized")
        if init:
            self._signals = GlobalSignals()
        return self._signals

    def args(self, args: Namespace = None) -> Optional[Namespace]:
        if self._args is None and args is None:
            raise RuntimeError("Uninitialized use of ArgumentsSingleton")
        if self._args is not None and args is not None:
            raise RuntimeError("Arguments already initialized")
        if args is not None:
            self._args = args
        return self._args

    def core(self, init: bool = False) -> LegendaryCore:
        if self._core is None and not init:
            raise RuntimeError("Uninitialized use of LegendaryCoreSingleton")
        if self._core is not None and init:
            raise RuntimeError("LegendaryCore already initialized")
        if init:
            try:
                self._core = LegendaryCore()
            except configparser.MissingSectionHeaderError as e:
                logger.warning(f"Config is corrupt: {e}")
                if config_path := os.environ.get("XDG_CONFIG_HOME"):
                    path = os.path.join(config_path, "legendary")
                else:
                    path = os.path.expanduser("~/.config/legendary")
                with open(os.path.join(path, "config.ini"), "w") as config_file:
                    config_file.write("[Legendary]")
                self._core = LegendaryCore()
            if "Legendary" not in self._core.lgd.config.sections():
                self._core.lgd.config.add_section("Legendary")
                self._core.lgd.save_config()
            # workaround if egl sync enabled, but no programdata_path
            # programdata_path might be unset if logging in through the browser
            if self._core.egl_sync_enabled:
                if self._core.egl.programdata_path is None:
                    self._core.lgd.config.remove_option("Legendary", "egl_sync")
                    self._core.lgd.save_config()
                else:
                    if not os.path.exists(self._core.egl.programdata_path):
                        self._core.lgd.config.remove_option("Legendary", "egl_sync")
                        self._core.lgd.save_config()
        return self._core

    def image_manager(self, init: bool = False) -> ImageManager:
        if self._image_manager is None and not init:
            raise RuntimeError("Uninitialized use of ImageManagerSingleton")
        if self._image_manager is not None and init:
            raise RuntimeError("ImageManager already initialized")
        if self._image_manager is None:
            self._image_manager = ImageManager(self.signals(), self.core())
        return self._image_manager

    def api_results(self, res: ApiResults = None) -> Optional[ApiResults]:
        if self._api_results is None and res is None:
            raise RuntimeError("Uninitialized use of ApiResultsSingleton")
        if self._api_results is not None and res is not None:
            raise RuntimeError("ApiResults already initialized")
        if res is not None:
            self._api_results = res
        return self._api_results

    def deleteLater(self) -> None:
        del self._api_results
        self._api_results = None

        self._image_manager.deleteLater()
        del self._image_manager
        self._image_manager = None

        self._core.exit()
        del self._core
        self._core = None

        self._signals.deleteLater()
        del self._signals
        self._signals = None

        del self._args
        self._args = None

        RareCore._instance = None

        super(RareCore, self).deleteLater()




import time
import webbrowser
from logging import getLogger

from PyQt5.QtCore import QObject, pyqtSignal, QRunnable, QThreadPool, QSize
from PyQt5.QtWidgets import QWidget, QLabel, QHBoxLayout, QSizePolicy, QPushButton, QGroupBox, QVBoxLayout

from legendary.models.game import Game
from rare.shared import LegendaryCoreSingleton, ArgumentsSingleton
from rare.utils.misc import icon

logger = getLogger("Ubisoft")


class UbiGetInfoSignals(QObject):
    worker_finished = pyqtSignal(set, set, str)


class UbiGetInfoWorker(QRunnable):
    def __init__(self):
        super(UbiGetInfoWorker, self).__init__()
        self.signals = UbiGetInfoSignals()
        self.setAutoDelete(True)
        self.core = LegendaryCoreSingleton()

    def run(self) -> None:
        try:
            external_auths = self.core.egs.get_external_auths()
            for ext_auth in external_auths:
                if ext_auth["type"] != "ubisoft":
                    continue
                ubi_account_id = ext_auth["externalAuthId"]
                break
            else:
                self.signals.worker_finished.emit(set(), set(), "")
                return

            uplay_keys = self.core.egs.store_get_uplay_codes()
            key_list = uplay_keys["data"]["PartnerIntegration"]["accountUplayCodes"]
            redeemed = {k["gameId"] for k in key_list if k["redeemedOnUplay"]}

            entitlements = self.core.egs.get_user_entitlements()
            entitlements = {i["entitlementName"] for i in entitlements}
            self.signals.worker_finished.emit(redeemed, entitlements, ubi_account_id)
        except Exception as e:
            logger.error(str(e))
            self.signals.worker_finished.emit(set(), set(), "error")


class UbiConnectSignals(QObject):
    linked = pyqtSignal(str)


class UbiConnectWorker(QRunnable):
    def __init__(self, ubi_account_id, partner_link_id):
        super(UbiConnectWorker, self).__init__()
        self.signals = UbiConnectSignals()
        self.setAutoDelete(True)
        self.core = LegendaryCoreSingleton()
        self.ubi_account_id = ubi_account_id
        self.partner_link_id = partner_link_id

    def run(self) -> None:
        if not self.ubi_account_id:  # debug
            time.sleep(2)
            self.signals.linked.emit("")
            return
        try:
            self.core.egs.store_claim_uplay_code(
                self.ubi_account_id, self.partner_link_id
            )
            self.core.egs.store_redeem_uplay_codes(self.ubi_account_id)
        except Exception as e:
            self.signals.linked.emit(str(e))
            return
        else:
            self.signals.linked.emit("")


class UbiLinkWidget(QWidget):
    def __init__(self, game: Game, ubi_account_id):
        super(UbiLinkWidget, self).__init__()
        self.args = ArgumentsSingleton()
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        self.game = game
        self.ubi_account_id = ubi_account_id

        self.title_label = QLabel(game.app_title)
        layout.addWidget(self.title_label, stretch=1)

        self.ok_indicator = QLabel()
        self.ok_indicator.setPixmap(icon("fa.info-circle", color="grey").pixmap(20, 20))
        self.ok_indicator.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Preferred)
        layout.addWidget(self.ok_indicator)

        self.link_button = QPushButton(
            self.tr("Redeem to Ubisoft") + ": Test" if self.args.debug else ""
        )
        layout.addWidget(self.link_button)
        self.link_button.clicked.connect(self.activate)
        self.setLayout(layout)

    def activate(self):
        self.link_button.setDisabled(True)
        # self.ok_indicator.setPixmap(icon("mdi.loading", color="grey").pixmap(20, 20))
        self.ok_indicator.setPixmap(
            icon("mdi.transit-connection-horizontal", color="grey").pixmap(20, 20)
        )

        if self.args.debug:
            worker = UbiConnectWorker(None, None)
        else:
            worker = UbiConnectWorker(self.ubi_account_id, self.game.partner_link_id)
        worker.signals.linked.connect(self.worker_finished)
        QThreadPool.globalInstance().start(worker)

    def worker_finished(self, error):
        if not error:
            self.ok_indicator.setPixmap(
                icon("ei.ok-circle", color="green").pixmap(QSize(20, 20))
            )
            self.link_button.setDisabled(True)
            self.link_button.setText(self.tr("Already activated"))
        else:
            self.ok_indicator.setPixmap(
                icon("fa.info-circle", color="red").pixmap(QSize(20, 20))
            )
            self.ok_indicator.setToolTip(error)
            self.link_button.setText(self.tr("Try again"))
            self.link_button.setDisabled(False)


class UbisoftGroup(QGroupBox):
    def __init__(self, parent=None):
        super(UbisoftGroup, self).__init__(parent=parent)
        self.setTitle(self.tr("Link Ubisoft Games"))
        self.setLayout(QVBoxLayout())
        self.core = LegendaryCoreSingleton()
        self.args = ArgumentsSingleton()

        self.thread_pool = QThreadPool.globalInstance()
        worker = UbiGetInfoWorker()
        worker.signals.worker_finished.connect(self.show_ubi_games)
        self.thread_pool.start(worker)

    def show_ubi_games(self, redeemed: set, entitlements: set, ubi_account_id: str):
        if not redeemed and ubi_account_id != "error":
            logger.error(
                "No linked ubisoft account found! Link your accounts via your browser and try again."
            )
            self.layout().addWidget(
                QLabel(
                    self.tr(
                        "Your account is not linked with Ubisoft. Please link your account first"
                    )
                )
            )
            open_browser_button = QPushButton(self.tr("Open link page"))
            open_browser_button.clicked.connect(
                lambda: webbrowser.open("https://www.epicgames.com/id/link/ubisoft")
            )
            self.layout().addWidget(open_browser_button)
            return
        elif ubi_account_id == "error":
            self.layout().addWidget(QLabel(self.tr("An error occurred")))
            return

        games = self.core.get_game_list(False)
        uplay_games = []
        activated = 0
        for game in games:
            for dlc_data in game.metadata.get("dlcItemList", []):
                if dlc_data["entitlementName"] not in entitlements:
                    continue

                try:
                    app_name = dlc_data["releaseInfo"][0]["appId"]
                except (IndexError, KeyError):
                    app_name = "unknown"

                dlc_game = Game(
                    app_name=app_name, app_title=dlc_data["title"], metadata=dlc_data
                )
                if dlc_game.partner_link_type != "ubisoft":
                    continue
                if dlc_game.partner_link_id in redeemed:
                    continue
                uplay_games.append(dlc_game)

            if game.partner_link_type != "ubisoft":
                continue
            if game.partner_link_id in redeemed:
                activated += 1
                continue
            uplay_games.append(game)

        if not uplay_games:
            if activated >= 1:
                self.layout().addWidget(
                    QLabel(
                        self.tr("All your Ubisoft games have already been activated")
                    )
                )
            else:
                self.layout().addWidget(
                    QLabel(self.tr("You don't own any Ubisoft games"))
                )
            if self.args.debug:
                widget = UbiLinkWidget(
                    Game(app_name="Test", app_title="This is a test game"),
                    ubi_account_id,
                )
                self.layout().addWidget(widget)
            return
        logger.info(f"Found {len(uplay_games)} game(s) to redeem")

        for game in uplay_games:
            widget = UbiLinkWidget(game, ubi_account_id)
            self.layout().addWidget(widget)



import os
from logging import getLogger
from typing import Callable, Tuple

from PyQt5.QtCore import (
    Qt,
    QCoreApplication,
    QRect,
    QSize,
    QPoint,
    pyqtSignal,
    QFileInfo,
)
from PyQt5.QtGui import QMovie, QPixmap, QFontMetrics, QImage
from PyQt5.QtWidgets import (
    QStyle,
    QSizePolicy,
    QLabel,
    QFileDialog,
    QHBoxLayout,
    QWidget,
    QPushButton,
    QStyleOptionTab,
    QStylePainter,
    QTabBar,
    QLineEdit,
    QToolButton,
    QTabWidget,
    QCompleter,
    QFileSystemModel,
    QStyledItemDelegate,
    QFileIconProvider,
    QVBoxLayout,
    QScrollArea,
)

from rare.utils.paths import tmp_dir
from rare.utils.qt_requests import QtRequestManager
from rare.utils.misc import icon as qta_icon

logger = getLogger("ExtraWidgets")


class IndicatorReasons:
    dir_not_empty = QCoreApplication.translate("IndicatorReasons", "Directory is not empty")
    wrong_format = QCoreApplication.translate("IndicatorReasons", "Given text has wrong format")
    game_not_installed = QCoreApplication.translate(
        "IndicatorReasons", "Game is not installed or does not exist"
    )
    dir_not_exist = QCoreApplication.translate("IndicatorReasons", "Directory does not exist")
    file_not_exist = QCoreApplication.translate("IndicatorReasons", "File does not exist")
    wrong_path = QCoreApplication.translate("IndicatorReasons", "Wrong Directory")


class IndicatorLineEdit(QWidget):
    textChanged = pyqtSignal(str)
    reasons = IndicatorReasons()

    def __init__(
        self,
        text: str = "",
        placeholder: str = "",
        completer: QCompleter = None,
        edit_func: Callable[[str], Tuple[bool, str, str]] = None,
        save_func: Callable[[str], None] = None,
        horiz_policy: QSizePolicy = QSizePolicy.Expanding,
        parent=None,
    ):
        super(IndicatorLineEdit, self).__init__(parent=parent)
        self.setObjectName(type(self).__name__)
        layout = QHBoxLayout(self)
        layout.setObjectName(f"{self.objectName()}Layout")
        layout.setContentsMargins(0, 0, 0, 0)
        # Add line_edit
        self.line_edit = QLineEdit(self)
        self.line_edit.setObjectName(f"{type(self).__name__}Edit")
        self.line_edit.setPlaceholderText(placeholder)
        self.line_edit.setSizePolicy(horiz_policy, QSizePolicy.Fixed)
        # Add hint_label to line_edit
        self.line_edit.setLayout(QHBoxLayout())
        self.hint_label = QLabel()
        self.hint_label.setObjectName(f"{type(self).__name__}Label")
        self.hint_label.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
        self.line_edit.layout().setContentsMargins(0, 0, 10, 0)
        self.line_edit.layout().addWidget(self.hint_label)
        # Add completer
        if completer is not None:
            completer.popup().setItemDelegate(QStyledItemDelegate(self))
            completer.popup().setAlternatingRowColors(True)
            self.line_edit.setCompleter(completer)
        layout.addWidget(self.line_edit)
        if edit_func is not None:
            self.indicator_label = QLabel()
            self.indicator_label.setPixmap(qta_icon("ei.info-circle", color="gray").pixmap(16, 16))
            self.indicator_label.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            layout.addWidget(self.indicator_label)

        if not placeholder:
            _translate = QCoreApplication.instance().translate
            self.line_edit.setPlaceholderText(_translate(self.__class__.__name__, "Default"))

        self.is_valid = False
        self.edit_func = edit_func
        self.save_func = save_func
        self.line_edit.textChanged.connect(self.__edit)
        if self.edit_func is None:
            self.line_edit.textChanged.connect(self.__save)

        # lk: this can be placed here to trigger __edit
        # lk: it is going to save the input again if it is valid which
        # lk: is ok to do given the checks don't misbehave (they shouldn't)
        # lk: however it is going to edit any "understood" bad input to good input
        # lk: and we might not want that (but the validity check reports on the edited string)
        # lk: it is also going to trigger this widget's textChanged signal but that gets lost
        if text:
            self.line_edit.setText(text)

    def text(self) -> str:
        return self.line_edit.text()

    def setText(self, text: str):
        self.line_edit.setText(text)

    def setHintText(self, text: str):
        self.hint_label.setFrameRect(self.line_edit.rect())
        self.hint_label.setText(text)

    def __indicator(self, res, reason=None):
        color = "green" if res else "red"
        self.indicator_label.setPixmap(qta_icon("ei.info-circle", color=color).pixmap(16, 16))
        if reason:
            self.indicator_label.setToolTip(reason)
        else:
            self.indicator_label.setToolTip("")

    def __edit(self, text):
        if self.edit_func is not None:
            self.line_edit.blockSignals(True)

            self.is_valid, text, reason = self.edit_func(text)
            if text != self.line_edit.text():
                self.line_edit.setText(text)
            self.line_edit.blockSignals(False)
            self.__indicator(self.is_valid, reason)
            if self.is_valid:
                self.__save(text)
            self.textChanged.emit(text)

    def __save(self, text):
        if self.save_func is not None:
            self.save_func(text)


class PathEditIconProvider(QFileIconProvider):
    icons = [
        ("mdi.file-cancel", "fa.file-excel-o"),  # Unknown
        ("mdi.desktop-classic", "fa.desktop"),  # Computer
        ("mdi.desktop-mac", "fa.desktop"),  # Desktop
        ("mdi.trash-can", "fa.trash"),  # Trashcan
        ("mdi.server-network", "fa.server"),  # Network
        ("mdi.harddisk", "fa.desktop"),  # Drive
        ("mdi.folder", "fa.folder"),  # Folder
        ("mdi.file", "fa.file"),  # File
        ("mdi.cog", "fa.cog"),  # Executable
    ]

    def __init__(self):
        super(PathEditIconProvider, self).__init__()
        self.icon_types = dict()
        for idx, (icn, fallback) in enumerate(self.icons):
            self.icon_types.update({idx - 1: qta_icon(icn, fallback, color="#eeeeee")})

    def icon(self, info_type):
        if isinstance(info_type, QFileInfo):
            if info_type.isRoot():
                return self.icon_types[4]
            if info_type.isDir():
                return self.icon_types[5]
            if info_type.isFile():
                return self.icon_types[6]
            if info_type.isExecutable():
                return self.icon_types[7]
            return self.icon_types[-1]
        return self.icon_types[int(info_type)]


class PathEdit(IndicatorLineEdit):
    def __init__(
        self,
        path: str = "",
        file_type: QFileDialog.FileType = QFileDialog.AnyFile,
        type_filter: str = "",
        name_filter: str = "",
        placeholder: str = "",
        edit_func: Callable[[str], Tuple[bool, str, str]] = None,
        save_func: Callable[[str], None] = None,
        horiz_policy: QSizePolicy = QSizePolicy.Expanding,
        parent=None,
    ):
        self.completer = QCompleter()
        self.compl_model = QFileSystemModel()

        try:
            self.compl_model.setOptions(
                QFileSystemModel.DontWatchForChanges
                | QFileSystemModel.DontResolveSymlinks
                | QFileSystemModel.DontUseCustomDirectoryIcons
            )
        except AttributeError as e:  # Error on Ubuntu
            logger.warning(e)
        self.compl_model.setIconProvider(PathEditIconProvider())
        self.compl_model.setRootPath(path)
        self.completer.setModel(self.compl_model)

        edit_func = self.__wrap_edit_function(edit_func)

        super(PathEdit, self).__init__(
            text=path,
            placeholder=placeholder,
            completer=self.completer,
            edit_func=edit_func,
            save_func=save_func,
            horiz_policy=horiz_policy,
            parent=parent,
        )
        self.setObjectName(type(self).__name__)
        self.line_edit.setMinimumSize(QSize(250, 0))
        self.path_select = QToolButton(self)
        self.path_select.setObjectName(f"{type(self).__name__}Button")
        layout = self.layout()
        layout.addWidget(self.path_select)

        _translate = QCoreApplication.instance().translate
        self.path_select.setText(_translate("PathEdit", "Browse..."))

        self.type_filter = type_filter
        self.name_filter = name_filter
        self.file_type = file_type

        self.path_select.clicked.connect(self.__set_path)

    def __set_path(self):
        dlg_path = self.line_edit.text()
        if not dlg_path:
            dlg_path = os.path.expanduser("~/")
        dlg = QFileDialog(self, self.tr("Choose path"), dlg_path)
        dlg.setFileMode(self.file_type)
        if self.type_filter:
            dlg.setFilter([self.type_filter])
        if self.name_filter:
            dlg.setNameFilter(self.name_filter)
        if dlg.exec_():
            names = dlg.selectedFiles()
            self.line_edit.setText(names[0])
            self.compl_model.setRootPath(names[0])

    def __wrap_edit_function(self, edit_function: Callable[[str], Tuple[bool, str, str]]):
        if edit_function:
            return lambda text: edit_function(os.path.expanduser(text)
                                              if text.startswith("~") else text)
        else:
            return edit_function


class SideTabBar(QTabBar):
    def __init__(self, parent=None):
        super(SideTabBar, self).__init__(parent=parent)
        self.setObjectName("SideTabBar")
        self.fm = QFontMetrics(self.font())

    def tabSizeHint(self, index):
        # width = QTabBar.tabSizeHint(self, index).width()
        return QSize(200, self.fm.height() + 18)

    def paintEvent(self, event):
        painter = QStylePainter(self)
        opt = QStyleOptionTab()

        for i in range(self.count()):
            self.initStyleOption(opt, i)
            painter.drawControl(QStyle.CE_TabBarTabShape, opt)
            painter.save()

            s = opt.rect.size()
            s.transpose()
            r = QRect(QPoint(), s)
            r.moveCenter(opt.rect.center())
            opt.rect = r

            c = self.tabRect(i).center()
            painter.translate(c)
            painter.rotate(90)
            painter.translate(-c)
            painter.drawControl(QStyle.CE_TabBarTabLabel, opt)
            painter.restore()


class SideTabContainer(QWidget):
    def __init__(self, widget: QWidget, title: str = str(), parent: QWidget = None):
        super(SideTabContainer, self).__init__(parent=parent)
        self.title = QLabel(self)
        self.setTitle(title)

        self.scroll = QScrollArea(self)
        self.scroll.setWidgetResizable(True)
        self.scroll.setSizeAdjustPolicy(QScrollArea.AdjustToContents)
        self.scroll.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.scroll.setFrameStyle(QScrollArea.NoFrame)
        if widget.layout():
            widget.layout().setAlignment(Qt.AlignTop)
            widget.layout().setContentsMargins(0, 0, 9, 0)
        widget.title = self.title
        widget.title.setTitle = self.setTitle
        self.scroll.setMinimumWidth(
            widget.sizeHint().width()
            + self.scroll.verticalScrollBar().sizeHint().width()
        )
        self.scroll.setWidget(widget)

        layout = QVBoxLayout()
        layout.addWidget(self.title)
        layout.addWidget(self.scroll)
        self.setLayout(layout)

    def setTitle(self, text: str) -> None:
        self.title.setText(f"<h2>{text}</h2>")
        self.title.setVisible(bool(text))


class SideTabWidget(QTabWidget):
    back_clicked = pyqtSignal()

    def __init__(self, show_back: bool = False, parent=None):
        super(SideTabWidget, self).__init__(parent=parent)
        self.setTabBar(SideTabBar())
        self.setDocumentMode(True)
        self.setTabPosition(QTabWidget.West)
        if show_back:
            super(SideTabWidget, self).addTab(
                QWidget(), qta_icon("mdi.keyboard-backspace", "ei.backward"), self.tr("Back")
            )
            self.tabBarClicked.connect(self.back_func)

    def back_func(self, tab):
        # shortcut for tab == 0
        if not tab:
            self.back_clicked.emit()

    def addTab(self, widget: QWidget, a1: str, title: str = str()) -> int:
        container = SideTabContainer(widget, title, parent=self)
        return super(SideTabWidget, self).addTab(container, a1)


class WaitingSpinner(QLabel):
    def __init__(self):
        super(WaitingSpinner, self).__init__()
        self.setStyleSheet(
            """
            margin-left: auto;
            margin-right: auto;
        """
        )
        self.movie = QMovie(":/images/loader.gif")
        self.setMovie(self.movie)
        self.movie.start()


class SelectViewWidget(QWidget):
    toggled = pyqtSignal()

    def __init__(self, icon_view: bool):
        super(SelectViewWidget, self).__init__()
        self.icon_view = icon_view
        self.setStyleSheet("""QPushButton{border: none; background-color: transparent}""")
        self.icon_button = QPushButton()
        self.list_button = QPushButton()
        if icon_view:
            self.icon_button.setIcon(
                qta_icon("mdi.view-grid-outline", "ei.th-large", color="orange")
            )
            self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list"))
        else:
            self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large"))
            self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="orange"))

        self.icon_button.clicked.connect(self.icon)
        self.list_button.clicked.connect(self.list)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.icon_button)
        layout.addWidget(self.list_button)

        self.setLayout(layout)

    def isChecked(self):
        return self.icon_view

    def icon(self):
        self.icon_button.setIcon(
            qta_icon("mdi.view-grid-outline", "ei.th-large", color="orange")
        )
        self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list"))
        self.icon_view = False
        self.toggled.emit()

    def list(self):
        self.icon_button.setIcon(qta_icon("mdi.view-grid-outline", "ei.th-large"))
        self.list_button.setIcon(qta_icon("fa5s.list", "ei.th-list", color="orange"))
        self.icon_view = True
        self.toggled.emit()


class ImageLabel(QLabel):
    image = None
    img_size = None
    name = ""

    def __init__(self):
        super(ImageLabel, self).__init__()
        self.path = tmp_dir()
        self.manager = QtRequestManager("bytes")

    def update_image(self, url, name="", size: tuple = (240, 320)):
        self.setFixedSize(*size)
        self.img_size = size
        self.name = name
        for c in r'<>?":|\/* ':
            self.name = self.name.replace(c, "")
        if self.img_size[0] > self.img_size[1]:
            name_extension = "wide"
        else:
            name_extension = "tall"
        self.name = f"{self.name}_{name_extension}.png"
        if not os.path.exists(os.path.join(self.path, self.name)):
            self.manager.get(url, self.image_ready)
            # self.request.finished.connect(self.image_ready)
        else:
            self.show_image()

    def image_ready(self, data):
        try:
            self.setPixmap(QPixmap())
        except Exception:
            logger.warning("C++ object already removed, when image ready")
            return
        image = QImage()
        image.loadFromData(data)
        image = image.scaled(
            *self.img_size[:2],
            Qt.KeepAspectRatio,
            transformMode=Qt.SmoothTransformation,
        )

        image.save(os.path.join(self.path, self.name))

        pixmap = QPixmap().fromImage(image)
        self.setPixmap(pixmap)

    def show_image(self):
        self.image = QPixmap(os.path.join(self.path, self.name)).scaled(
            *self.img_size, transformMode=Qt.SmoothTransformation
        )
        self.setPixmap(self.image)


class ButtonLineEdit(QLineEdit):
    buttonClicked = pyqtSignal()

    def __init__(self, icon_name, placeholder_text: str, parent=None):
        super(ButtonLineEdit, self).__init__(parent)

        self.button = QToolButton(self)
        self.button.setIcon(qta_icon(icon_name, color="white"))
        self.button.setStyleSheet("border: 0px; padding: 0px;")
        self.button.setCursor(Qt.ArrowCursor)
        self.button.clicked.connect(self.buttonClicked.emit)
        self.setPlaceholderText(placeholder_text)
        frameWidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        buttonSize = self.button.sizeHint()

        self.setStyleSheet(
            "QLineEdit {padding-right: %dpx; }" % (buttonSize.width() + frameWidth + 1)
        )
        self.setMinimumSize(
            max(self.minimumSizeHint().width(), buttonSize.width() + frameWidth * 2 + 2),
            max(
                self.minimumSizeHint().height(),
                buttonSize.height() + frameWidth * 2 + 2,
            ),
        )

    def resizeEvent(self, event):
        buttonSize = self.button.sizeHint()
        frameWidth = self.style().pixelMetric(QStyle.PM_DefaultFrameWidth)
        self.button.move(
            self.rect().right() - frameWidth - buttonSize.width(),
            (self.rect().bottom() - buttonSize.height() + 1) // 2,
        )
        super(ButtonLineEdit, self).resizeEvent(event)
